import logging
from pathlib import Path

import whoosh
from whoosh import qparser
from whoosh.filedb.filestore import FileStorage
from whoosh.index import EmptyIndexError
from whoosh.query import Every

from django.conf import settings

from mayan.apps.common.utils import any_to_bool, parse_range
from mayan.apps.lock_manager.backends.base import LockingBackend
from mayan.apps.lock_manager.exceptions import LockError

from ..classes import SearchBackend, SearchModel
from ..exceptions import DynamicSearchRetry
from ..settings import setting_results_limit

from .literals import (
    DJANGO_TO_WHOOSH_FIELD_MAP, TEXT_LOCK_INSTANCE_DEINDEX,
    TEXT_LOCK_INSTANCE_INDEX, WHOOSH_INDEX_DIRECTORY_NAME,
)
logger = logging.getLogger(name=__name__)


class WhooshSearchBackend(SearchBackend):
    field_map = DJANGO_TO_WHOOSH_FIELD_MAP

    def __init__(self, **kwargs):
        index_path = kwargs.pop('index_path', None)
        writer_limitmb = kwargs.pop('writer_limitmb', 128)
        writer_multisegment = kwargs.pop('writer_multisegment', False)
        writer_procs = kwargs.pop('writer_procs', 1)

        super().__init__(**kwargs)

        self.index_path = Path(
            index_path or Path(settings.MEDIA_ROOT, WHOOSH_INDEX_DIRECTORY_NAME)
        )

        if writer_limitmb:
            writer_limitmb = int(writer_limitmb)

        if writer_multisegment:
            writer_multisegment = any_to_bool(value=writer_multisegment)

        if writer_procs:
            writer_procs = int(writer_procs)

        self.writer_kwargs = {
            'limitmb': writer_limitmb, 'multisegment': writer_multisegment,
            'procs': writer_procs
        }

    def _get_status(self):
        result = []

        title = 'Whoosh search model indexing status'
        result.append(title)
        result.append(len(title) * '=')

        for search_model in SearchModel.all():
            index = self.get_or_create_index(search_model=search_model)
            search_results = index.searcher().search(Every('id'))

            result.append(
                '{}: {}'.format(
                    search_model.label, search_results.estimated_length()
                )
            )

        return '\n'.join(result)

    def _initialize(self):
        self.index_path.mkdir(exist_ok=True)

    def _search(
        self, query, search_model, user, global_and_search=False,
        ignore_limit=False
    ):
        index = self.get_or_create_index(search_model=search_model)

        id_list = []
        with index.searcher() as searcher:
            search_string = []

            for key, value in query.items():
                search_string.append(
                    '{}:({})'.format(key, value)
                )

            global_logic_string = ' AND ' if global_and_search else ' OR '
            search_string = global_logic_string.join(search_string)

            logger.debug('search_string: %s', search_string)

            parser = qparser.QueryParser(
                fieldname='_', schema=index.schema
            )
            parser.remove_plugin_class(cls=qparser.WildcardPlugin)
            parser.add_plugin(pin=qparser.PrefixPlugin())
            whoosh_query = parser.parse(text=search_string)

            if ignore_limit:
                limit = None
            else:
                limit = setting_results_limit.value

            results = searcher.search(q=whoosh_query, limit=limit)

            logger.debug('results: %s', results)

            for result in results:
                id_list.append(result['id'])

        return search_model.get_queryset().filter(
            id__in=id_list
        ).distinct()

    def clear_search_model_index(self, search_model):
        schema = self.get_search_model_schema(search_model=search_model)

        # Clear the model index.
        self.get_storage().create_index(
            indexname=search_model.get_full_name(), schema=schema
        )

    def deindex_instance(self, instance):
        try:
            lock = LockingBackend.get_backend().acquire_lock(
                name=TEXT_LOCK_INSTANCE_DEINDEX
            )
        except LockError:
            raise
        else:
            try:
                search_model = SearchModel.get_for_model(instance=instance)
                index = self.get_or_create_index(search_model=search_model)

                with index.writer(**self.writer_kwargs) as writer:
                    writer.delete_by_term('id', str(instance.pk))
            finally:
                lock.release()

    def get_or_create_index(self, search_model):
        storage = self.get_storage()
        schema = self.get_search_model_schema(search_model=search_model)

        try:
            # Explicitly specify the schema. Allows using existing index
            # when the schema changes.
            index = storage.open_index(
                indexname=search_model.get_full_name(), schema=schema
            )
        except EmptyIndexError:
            index = storage.create_index(
                indexname=search_model.get_full_name(), schema=schema
            )

        return index

    def get_search_model_schema(self, search_model):
        field_map = self.get_resolved_field_map(search_model=search_model)
        schema_kwargs = {key: value['field'] for key, value in field_map.items()}

        return whoosh.fields.Schema(**schema_kwargs)

    def get_storage(self):
        return FileStorage(path=self.index_path)

    def index_instance(self, instance, exclude_model=None, exclude_kwargs=None):
        try:
            lock = LockingBackend.get_backend().acquire_lock(
                name=TEXT_LOCK_INSTANCE_INDEX
            )
        except LockError:
            raise
        else:
            try:
                search_model = SearchModel.get_for_model(instance=instance)
                index = self.get_or_create_index(search_model=search_model)

                with index.writer(**self.writer_kwargs) as writer:
                    kwargs = search_model.populate(
                        backend=self, instance=instance,
                        exclude_model=exclude_model,
                        exclude_kwargs=exclude_kwargs
                    )

                    try:
                        writer.delete_by_term('id', str(instance.pk))
                        writer.add_document(**kwargs)
                    except Exception as exception:
                        logger.error(
                            'Unexpected exception while indexing object '
                            'id: %(id)s, search model: %(search_model)s, '
                            'index data: %(index_data)s, raw data: '
                            '%(raw_data)s, field map: %(field_map)s; '
                            '%(exception)s' % {
                                'exception': exception,
                                'field_map': self.get_resolved_field_map(
                                    search_model=search_model
                                ),
                                'id': instance.pk,
                                'index_data': kwargs,
                                'raw_data': instance.__dict__,
                                'search_model': search_model.get_full_name()
                            }, exc_info=True
                        )
                        raise
            except whoosh.index.LockError:
                raise DynamicSearchRetry
            finally:
                lock.release()

    def index_search_model(self, search_model, range_string=None):
        queryset = search_model.get_queryset()

        if range_string:
            queryset = queryset.filter(
                pk__in=list(parse_range(range_string=range_string))
            )

        for instance in queryset:
            self.index_instance(instance=instance)

    def reset(self, search_model=None):
        self.tear_down(search_model=search_model)
        self.update_mappings(search_model=search_model)

    def tear_down(self, search_model=None):
        if search_model:
            search_models = (search_model,)
        else:
            search_models = SearchModel.all()

        for search_model in search_models:
            self.clear_search_model_index(search_model=search_model)

    def update_mappings(self, search_model=None):
        if search_model:
            search_models = (search_model,)
        else:
            search_models = SearchModel.all()

        for search_model in search_models:
            self.get_or_create_index(search_model=search_model)



from django.db import models, migrations

import mayan.apps.common.validators


class Migration(migrations.Migration):
    dependencies = [
        ('converter', '0002_auto_20150608_1943'),
    ]

    operations = [
        migrations.AlterField(
            model_name='transformation',
            name='arguments',
            field=models.TextField(
                default='', help_text='Enter the arguments for the '
                'transformation as a YAML dictionary. ie: {"degrees": 180}',
                blank=True, verbose_name='Arguments', validators=[
                    mayan.apps.common.validators.YAMLValidator
                ]
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='transformation',
            name='name',
            field=models.CharField(
                max_length=128, verbose_name='Name', choices=[
                    ('rotate', 'Rotate <degrees>'), ('zoom', 'Zoom <percent>'),
                    ('resize', 'Resize <width, height>')
                ]
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='transformation',
            name='order',
            field=models.PositiveIntegerField(
                default=0, blank=True, help_text='Order in which the '
                'transformations will be executed.', null=True,
                verbose_name='Order', db_index=True
            ),
            preserve_default=True,
        ),
    ]



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from .mixins import CommonAPITestMixin


class CommonAPITestCase(CommonAPITestMixin, BaseAPITestCase):
    auto_login_user = False

    def test_content_type_list_api_view(self):
        response = self._request_content_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)



from django.db.models import Q

from ..models import DocumentMetadata, DocumentTypeMetadataType, MetadataType

from .literals import (
    TEST_METADATA_TYPE_LABEL, TEST_METADATA_TYPE_LABEL_EDITED,
    TEST_METADATA_TYPE_NAME, TEST_METADATA_TYPE_NAME_EDITED,
    TEST_METADATA_VALUE, TEST_METADATA_VALUE_EDITED
)


class DocumentMetadataAPIViewTestMixin:
    def _request_document_metadata_create_api_view(self, extra_data=None):
        pk_list = list(DocumentMetadata.objects.values_list('pk', flat=True))

        data = {
            'metadata_type_id': self.test_metadata_type.pk,
            'value': TEST_METADATA_VALUE
        }
        data.update(extra_data or {})

        response = self.post(
            viewname='rest_api:documentmetadata-list',
            kwargs={'document_id': self.test_document.pk}, data=data
        )

        try:
            self.test_document_metadata = DocumentMetadata.objects.get(
                ~Q(pk__in=pk_list)
            )
        except DocumentMetadata.DoesNotExist:
            self.test_document_metadata = None

        return response

    def _request_document_metadata_delete_api_view(self):
        return self.delete(
            viewname='rest_api:documentmetadata-detail',
            kwargs={
                'document_id': self.test_document.pk,
                'metadata_id': self.test_document_metadata.pk
            }
        )

    def _request_document_metadata_edit_api_view_via_patch(self, extra_data=None):
        data = {
            'value': TEST_METADATA_VALUE_EDITED
        }
        data.update(extra_data or {})

        return self.patch(
            viewname='rest_api:documentmetadata-detail',
            kwargs={
                'document_id': self.test_document.pk,
                'metadata_id': self.test_document_metadata.pk
            }, data=data
        )

    def _request_document_metadata_edit_api_view_via_put(self, extra_data=None):
        data = {
            'value': TEST_METADATA_VALUE_EDITED
        }
        data.update(extra_data or {})

        return self.put(
            viewname='rest_api:documentmetadata-detail',
            kwargs={
                'document_id': self.test_document.pk,
                'metadata_id': self.test_document_metadata.pk
            }, data=data
        )

    def _request_document_metadata_list_api_view(self):
        return self.get(
            viewname='rest_api:documentmetadata-list', kwargs={
                'document_id': self.test_document.pk
            }
        )


class DocumentMetadataMixin:
    def _create_test_document_metadata(self):
        self.test_document_metadata = self.test_document.metadata.create(
            metadata_type=self.test_metadata_type, value=TEST_METADATA_VALUE
        )


class DocumentMetadataViewTestMixin:
    def _request_test_document_metadata_add_get_view(self):
        return self.get(
            viewname='metadata:metadata_add', kwargs={
                'document_id': self.test_document.pk
            }, data={'metadata_type': self.test_metadata_type.pk}
        )

    def _request_test_document_metadata_add_post_view(self):
        return self.post(
            viewname='metadata:metadata_add', kwargs={
                'document_id': self.test_document.pk
            }, data={'metadata_type': self.test_metadata_type.pk}
        )

    def _request_test_document_metadata_multiple_add_post_view(self):
        return self.post(
            viewname='metadata:metadata_add', kwargs={
                'document_id': self.test_document.pk
            }, data={
                'metadata_type': [
                    metadata_type.pk for metadata_type in self.test_metadata_types
                ],
            }
        )

    def _request_test_document_metadata_edit_post_view(
        self, extra_data=None, follow=False
    ):
        data = {
            'form-0-metadata_type_id': self.test_metadata_type.pk,
            'form-0-update': True,
            'form-0-value': TEST_METADATA_VALUE_EDITED,
            'form-TOTAL_FORMS': '1',
            'form-INITIAL_FORMS': '0',
            'form-MAX_NUM_FORMS': '',
        }

        if extra_data:
            data.update(extra_data)

        return self.post(
            viewname='metadata:metadata_edit', kwargs={
                'document_id': self.test_document.pk
            }, data=data, follow=follow
        )

    def _request_test_document_metadata_list_view(self):
        return self.get(
            viewname='metadata:metadata_view', kwargs={
                'document_id': self.test_document.pk
            }
        )

    def _request_test_document_metadata_remove_get_view(self):
        return self.get(
            viewname='metadata:metadata_remove', kwargs={
                'document_id': self.test_document.pk
            }
        )

    def _request_test_document_metadata_remove_post_view(self, index=0):
        return self.post(
            viewname='metadata:metadata_remove',
            kwargs={'document_id': self.test_document.pk}, data={
                'form-0-metadata_type_id': self.test_metadata_types[index].pk,
                'form-0-update': True,
                'form-TOTAL_FORMS': '1',
                'form-INITIAL_FORMS': '0',
                'form-MAX_NUM_FORMS': '',
            }
        )

    def _request_test_document_multiple_metadata_add_post_view(self):
        return self.post(
            viewname='metadata:metadata_multiple_add', data={
                'id_list': '{},{}'.format(
                    self.test_documents[0].pk, self.test_documents[1].pk
                ),
                'metadata_type': self.test_metadata_type.pk
            }
        )

    def _request_test_document_multiple_metadata_edit_get_view(self):
        return self.get(
            viewname='metadata:metadata_multiple_edit', data={
                'id_list': '{},{}'.format(
                    self.test_documents[0].pk, self.test_documents[1].pk
                )
            }
        )

    def _request_test_document_multiple_metadata_edit_post_view(self):
        return self.post(
            viewname='metadata:metadata_multiple_edit', data={
                'id_list': '{},{}'.format(
                    self.test_documents[0].pk, self.test_documents[1].pk
                ),
                'form-0-metadata_type_id': self.test_metadata_type.pk,
                'form-0-update': True,
                'form-0-value': TEST_METADATA_VALUE_EDITED,
                'form-TOTAL_FORMS': '1',
                'form-INITIAL_FORMS': '0',
                'form-MAX_NUM_FORMS': '',
            }
        )

    def _request_test_document_multiple_metadata_remove_get_view(self):
        return self.get(
            viewname='metadata:metadata_multiple_remove', data={
                'id_list': '{},{}'.format(
                    self.test_documents[0].pk, self.test_documents[0].pk
                )
            }
        )

    def _request_test_document_multiple_metadata_remove_post_view(self):
        return self.post(
            viewname='metadata:metadata_multiple_remove', data={
                'id_list': '{},{}'.format(
                    self.test_documents[0].pk, self.test_documents[1].pk
                ),
                'form-0-metadata_type_id': self.test_metadata_type.pk,
                'form-0-update': True,
                'form-TOTAL_FORMS': '1',
                'form-INITIAL_FORMS': '0',
                'form-MAX_NUM_FORMS': '',
            }
        )


class DocumentTypeMetadataTypeAPIViewTestMixin:
    def _request_document_type_metadata_type_create_api_view(self):
        pk_list = list(DocumentTypeMetadataType.objects.values_list('pk', flat=True))

        response = self.post(
            viewname='rest_api:documenttypemetadatatype-list',
            kwargs={'document_type_id': self.test_document_type.pk}, data={
                'metadata_type_id': self.test_metadata_type.pk,
                'required': False
            }
        )

        try:
            self.test_document_type_metadata_type = DocumentTypeMetadataType.objects.get(
                ~Q(pk__in=pk_list)
            )
        except DocumentTypeMetadataType.DoesNotExist:
            self.test_document_type_metadata_type = None

        return response

    def _request_document_type_metadata_type_delete_api_view(self):
        return self.delete(
            viewname='rest_api:documenttypemetadatatype-detail',
            kwargs={
                'document_type_id': self.test_document_type.pk,
                'metadata_type_id': self.test_document_type_metadata_type.pk
            }
        )

    def _request_document_type_metadata_type_list_api_view(self):
        return self.get(
            viewname='rest_api:documenttypemetadatatype-list', kwargs={
                'document_type_id': self.test_document_type.pk
            }
        )

    def _request_document_type_metadata_type_edit_api_view_via_patch(self):
        return self.patch(
            viewname='rest_api:documenttypemetadatatype-detail',
            kwargs={
                'document_type_id': self.test_document_type.pk,
                'metadata_type_id': self.test_document_type_metadata_type.pk
            }, data={
                'required': True
            }
        )

    def _request_document_type_metadata_type_edit_api_view_via_put(self):
        return self.put(
            viewname='rest_api:documenttypemetadatatype-detail',
            kwargs={
                'document_type_id': self.test_document_type.pk,
                'metadata_type_id': self.test_document_type_metadata_type.pk
            }, data={
                'required': True
            }
        )


class DocumentTypeMetadataTypeTestMixin:
    def _create_test_document_type_metadata_type(self):
        self.test_document_type_metadata_type = self.test_document_type.metadata.create(
            metadata_type=self.test_metadata_type, required=False
        )


class MetadataTypeAPIViewTestMixin:
    def _request_test_metadata_type_create_api_view(self):
        pk_list = list(MetadataType.objects.values('pk'))

        response = self.post(
            viewname='rest_api:metadatatype-list', data={
                'name': 'test_metadata_type', 'label': 'test metadata type'
            }
        )

        try:
            self.test_metadata_type = MetadataType.objects.get(
                ~Q(pk__in=pk_list)
            )
        except MetadataType.DoesNotExist:
            self.test_metadata_type = None

        return response

    def _request_test_metadata_type_delete_api_view(self):
        return self.delete(
            viewname='rest_api:metadatatype-detail',
            kwargs={'metadata_type_id': self.test_metadata_type.pk}
        )

    def _request_test_metadata_type_detail_api_view(self):
        return self.get(
            viewname='rest_api:metadatatype-detail',
            kwargs={'metadata_type_id': self.test_metadata_type.pk}
        )

    def _request_test_metadata_type_edit_api_view_via_patch(self):
        return self.patch(
            viewname='rest_api:metadatatype-detail',
            kwargs={'metadata_type_id': self.test_metadata_type.pk}, data={
                'label': '{} edited'.format(self.test_metadata_type.label),
                'name': '{}_edited'.format(self.test_metadata_type.name),
            }
        )

    def _request_test_metadata_type_edit_api_view_via_put(self):
        return self.put(
            viewname='rest_api:metadatatype-detail',
            kwargs={'metadata_type_id': self.test_metadata_type.pk}, data={
                'label': '{} edited'.format(self.test_metadata_type.label),
                'name': '{}_edited'.format(self.test_metadata_type.name),
            }
        )

    def _request_test_metadata_type_list_api_view(self):
        return self.get(viewname='rest_api:metadatatype-list')


class MetadataTypeTestMixin:
    def setUp(self):
        super().setUp()
        self.test_metadata_types = []
        self._test_document_type_metadata_type_relationships = []

    def _get_test_metadata_type_queryset(self):
        return MetadataType.objects.filter(
            pk__in=[
                metadata_type.pk for metadata_type in self.test_metadata_types
            ]
        )

    def _create_test_metadata_type(
        self, add_test_document_type=False, extra_kwargs=None, required=False
    ):
        total_test_metadata_types = len(self.test_metadata_types)
        name = '{}_{}'.format(
            TEST_METADATA_TYPE_NAME, total_test_metadata_types
        )
        label = '{}_{}'.format(
            TEST_METADATA_TYPE_LABEL, total_test_metadata_types
        )

        kwargs = {'name': name, 'label': label}

        if extra_kwargs:
            kwargs.update(extra_kwargs)

        self.test_metadata_type = MetadataType.objects.create(**kwargs)
        self.test_metadata_types.append(self.test_metadata_type)

        if add_test_document_type:
            self._test_document_type_metadata_type_relationships.append(
                self.test_document_type.metadata.create(
                    metadata_type=self.test_metadata_type, required=required
                )
            )


class MetadataTypeViewTestMixin:
    def _request_test_document_type_relationship_delete_view(self):
        # This request assumes there is only one document type and
        # blindly sets the first form of the formset.

        return self.post(
            viewname='metadata:document_type_metadata_type_relationship',
            kwargs={'document_type_id': self.test_document_type.pk}, data={
                'form-TOTAL_FORMS': '1',
                'form-INITIAL_FORMS': '0',
                'form-0-relationship_type': 'none'
            }
        )

    def _request_test_document_type_relationship_edit_view(self):
        # This request assumes there is only one document type and
        # blindly sets the first form of the formset.

        return self.post(
            viewname='metadata:document_type_metadata_type_relationship',
            kwargs={'document_type_id': self.test_document_type.pk}, data={
                'form-TOTAL_FORMS': '1',
                'form-INITIAL_FORMS': '0',
                'form-0-relationship_type': 'required'
            }
        )

    def _request_test_metadata_type_create_view(self):
        pk_list = list(MetadataType.objects.values('pk'))

        response = self.post(
            viewname='metadata:metadata_type_create', data={
                'name': 'test_metadata_type', 'label': 'test metadata type'
            }
        )

        try:
            self.test_metadata_type = MetadataType.objects.get(
                ~Q(pk__in=pk_list)
            )
        except MetadataType.DoesNotExist:
            self.test_metadata_type = None

        return response

    def _request_test_metadata_type_delete_single_view(self):
        return self.post(
            viewname='metadata:metadata_type_delete_single', kwargs={
                'metadata_type_id': self.test_metadata_type.pk
            }
        )

    def _request_test_metadata_type_delete_multiple_view(self):
        return self.post(
            viewname='metadata:metadata_type_delete_multiple', data={
                'id_list': self.test_metadata_type.pk
            }
        )

    def _request_test_metadata_type_edit_view(self):
        return self.post(
            viewname='metadata:metadata_type_edit', kwargs={
                'metadata_type_id': self.test_metadata_type.pk
            }, data={
                'label': TEST_METADATA_TYPE_LABEL_EDITED,
                'name': TEST_METADATA_TYPE_NAME_EDITED
            }
        )

    def _request_metadata_type_list_view(self):
        return self.get(
            viewname='metadata:metadata_type_list',
        )

    def _request_test_metadata_type_document_type_relationship_delete_view(self):
        # This request assumes there is only one document type and
        # blindly sets the first form of the formset.

        return self.post(
            viewname='metadata:metadata_type_document_type_relationship',
            kwargs={'metadata_type_id': self.test_metadata_type.pk}, data={
                'form-TOTAL_FORMS': '1',
                'form-INITIAL_FORMS': '0',
                'form-0-relationship_type': 'none'
            }
        )

    def _request_test_metadata_type_document_type_relationship_edit_view(self):
        # This request assumes there is only one document type and
        # blindly sets the first form of the formset.

        return self.post(
            viewname='metadata:metadata_type_document_type_relationship',
            kwargs={'metadata_type_id': self.test_metadata_type.pk}, data={
                'form-TOTAL_FORMS': '1',
                'form-INITIAL_FORMS': '0',
                'form-0-relationship_type': 'required'
            }
        )



import logging

from django.apps import apps
from django.contrib.admin.utils import (
    get_fields_from_path, reverse_field_path
)
from django.db.models.signals import m2m_changed, post_save, pre_delete
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.translation import ugettext as _

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.utils import (
    ResolverPipelineModelAttribute, flatten_list, get_class_full_name
)
from mayan.apps.views.literals import LIST_MODE_CHOICE_LIST

from .exceptions import DynamicSearchException
from .literals import (
    DEFAULT_SCOPE_ID, DELIMITER, MESSAGE_FEATURE_NO_STATUS,
    QUERY_PARAMETER_ANY_FIELD, SCOPE_MATCH_ALL,
    SCOPE_MARKER, SCOPE_OPERATOR_CHOICES, SCOPE_OPERATOR_MARKER,
    SCOPE_RESULT_MAKER
)
from .settings import (
    setting_backend, setting_backend_arguments,
    setting_results_limit
)
logger = logging.getLogger(name=__name__)


class SearchBackend:
    _initialized = False
    _search_field_transformations = {}

    @staticmethod
    def _disable():
        for search_model in SearchModel.all():
            post_save.disconnect(
                dispatch_uid='search_handler_index_instance',
                sender=search_model.model
            )
            pre_delete.disconnect(
                dispatch_uid='search_handler_deindex_instance',
                sender=search_model.model
            )

            for proxy in search_model.proxies:
                post_save.disconnect(
                    dispatch_uid='search_handler_index_instance',
                    sender=proxy
                )
                pre_delete.disconnect(
                    dispatch_uid='search_handler_deindex_instance',
                    sender=proxy
                )

            for related_model, path in search_model.get_related_models():
                post_save.disconnect(
                    dispatch_uid='search_handler_index_related_instance_{}_{}'.format(
                        get_class_full_name(klass=search_model.model),
                        get_class_full_name(klass=related_model)
                    ), sender=related_model
                )
                pre_delete.disconnect(
                    dispatch_uid='search_handler_index_related_instance_delete_{}_{}'.format(
                        get_class_full_name(klass=search_model.model),
                        get_class_full_name(klass=related_model)
                    ), sender=related_model
                )

        for through_model, data in SearchModel.get_through_models().items():
            m2m_changed.disconnect(
                dispatch_uid='search_handler_index_related_instance_m2m_{}'.format(
                    get_class_full_name(klass=through_model),
                ), sender=through_model
            )

    @staticmethod
    def _enable():
        # Hidden import.
        from .handlers import (
            handler_deindex_instance, handler_index_instance,
            handler_factory_index_related_instance_delete,
            handler_factory_index_related_instance_m2m,
            handler_factory_index_related_instance_save
        )

        for search_model in SearchModel.all():
            post_save.connect(
                dispatch_uid='search_handler_index_instance',
                receiver=handler_index_instance, sender=search_model.model
            )
            pre_delete.connect(
                dispatch_uid='search_handler_deindex_instance',
                receiver=handler_deindex_instance,
                sender=search_model.model, weak=False
            )

            for proxy in search_model.proxies:
                post_save.connect(
                    dispatch_uid='search_handler_index_instance',
                    receiver=handler_index_instance, sender=proxy
                )
                pre_delete.connect(
                    dispatch_uid='search_handler_deindex_instance',
                    receiver=handler_deindex_instance,
                    sender=proxy, weak=False
                )

            for related_model, path in search_model.get_related_models():
                post_save.connect(
                    dispatch_uid='search_handler_index_related_instance_{}_{}'.format(
                        get_class_full_name(klass=search_model.model),
                        get_class_full_name(klass=related_model)
                    ),
                    receiver=handler_factory_index_related_instance_save(
                        reverse_field_path=path
                    ), sender=related_model, weak=False
                )
                pre_delete.connect(
                    dispatch_uid='search_handler_index_related_instance_delete_{}_{}'.format(
                        get_class_full_name(klass=search_model.model),
                        get_class_full_name(klass=related_model)
                    ),
                    receiver=handler_factory_index_related_instance_delete(
                        reverse_field_path=path
                    ), sender=related_model, weak=False
                )

        for through_model, data in SearchModel.get_through_models().items():
            m2m_changed.connect(
                dispatch_uid='search_handler_index_related_instance_m2m_{}'.format(
                    get_class_full_name(klass=through_model)
                ),
                receiver=handler_factory_index_related_instance_m2m(
                    data=data
                ), sender=through_model, weak=False
            )

    @staticmethod
    def get_instance(extra_kwargs=None):
        kwargs = setting_backend_arguments.value.copy()
        if extra_kwargs:
            kwargs.update(extra_kwargs)

        return import_string(dotted_path=setting_backend.value)(
            **kwargs
        )

    @staticmethod
    def limit_queryset(queryset):
        pk_list = queryset.values('pk')[:setting_results_limit.value]
        return queryset.filter(pk__in=pk_list)

    @staticmethod
    def index_related_instance_m2m(
        action, instance, model, pk_set, search_model_related_paths
    ):
        # Hidden import
        from .tasks import task_index_instance

        if action in ('post_add', 'pre_remove'):
            instance_paths = search_model_related_paths.get(instance._meta.model, ())
            model_paths = search_model_related_paths.get(model, ())

            if action == 'pre_remove':
                exclude_kwargs = {
                    'exclude_app_label': instance._meta.app_label,
                    'exclude_model_name': instance._meta.model_name,
                    'exclude_kwargs': {'id': instance.pk}
                }
            else:
                exclude_kwargs = {}

            for instance_path in instance_paths:
                result = ResolverPipelineModelAttribute.resolve(
                    attribute=instance_path, obj=instance
                )

                entries = flatten_list(value=result)

                for entry in entries:
                    task_kwargs = {
                        'app_label': entry._meta.app_label,
                        'model_name': entry._meta.model_name,
                        'object_id': entry.pk
                    }
                    task_kwargs.update(exclude_kwargs)

                    task_index_instance.apply_async(
                        kwargs=task_kwargs
                    )

            if action == 'pre_remove':
                exclude_kwargs = {
                    'exclude_app_label': model._meta.app_label,
                    'exclude_model_name': model._meta.model_name,
                    'exclude_kwargs': {'id__in': pk_set}
                }
            else:
                exclude_kwargs = {}

            for model_instance in model._meta.default_manager.filter(pk__in=pk_set):
                for instance_path in model_paths:
                    result = ResolverPipelineModelAttribute.resolve(
                        attribute=instance_path, obj=model_instance
                    )

                    entries = flatten_list(value=result)

                    for entry in entries:
                        task_kwargs = {
                            'app_label': entry._meta.app_label,
                            'model_name': entry._meta.model_name,
                            'object_id': entry.pk
                        }
                        task_kwargs.update(exclude_kwargs)

                        task_index_instance.apply_async(
                            kwargs=task_kwargs
                        )

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def _search(self, global_and_search, query, search_model, user, ignore_limit):
        raise NotImplementedError

    def cleanup_query(self, query, search_model):
        search_field_names = [
            search_field.get_full_name() for search_field in search_model.search_fields
        ]

        clean_query = {}

        if QUERY_PARAMETER_ANY_FIELD in query:
            value = query[QUERY_PARAMETER_ANY_FIELD]
            if value:
                clean_query = {key: value for key in search_field_names}
        else:
            # Allow only valid search fields for the search model and scoping keys.
            clean_query = {
                key: value for key, value in query.items() if key in search_field_names and value
            }

        return clean_query

    def close(self):
        """
        Optional method to terminate a backend instance, such as closing
        connections.
        """

    def decode_query(self, query, global_and_search=False):
        # Clean up the query.
        query.pop('_match_all', None)

        # Turn scoped query dictionary into a series of unscoped queries.
        operators = {}
        result_scope = DEFAULT_SCOPE_ID
        scope_match_all = False
        scopes = {}

        for key, value in query.items():
            scope_id = DEFAULT_SCOPE_ID

            # Check if the entry has a scope marker.
            if key.startswith(SCOPE_MARKER):
                # Remove the scope marker.
                key = key[len(SCOPE_MARKER):]

                if key.startswith(SCOPE_OPERATOR_MARKER):
                    # Check for operator.
                    # __operator_SCOPE_SCOPE=OPERATOR_SCOPE
                    key = key[len(SCOPE_OPERATOR_MARKER):]
                    operator_scopes = key[len(DELIMITER):].split(DELIMITER)
                    operator_text, result = value.split(DELIMITER)

                    operators[result] = {
                        'scopes': operator_scopes,
                        'function': SCOPE_OPERATOR_CHOICES[operator_text],
                    }
                elif key.startswith(SCOPE_RESULT_MAKER):
                    # Check for result.
                    # __result=SCOPE
                    result_scope = value
                else:
                    # Check scope match all.
                    # __SCOPE_match_all
                    if key.endswith(SCOPE_MATCH_ALL):
                        scope_id, key = key.split(DELIMITER, 1)
                        scopes.setdefault(scope_id, {})
                        scope_match_all = value.lower() in ['on', 'true']
                        scopes[scope_id]['match_all'] = scope_match_all
                    else:
                        # Must be a scoped query.
                        # __SCOPE_QUERY=VALUE
                        scope_id, key = key.split(DELIMITER, 1)
                        scopes.setdefault(scope_id, {})
                        scopes[scope_id].setdefault('match_all', False)
                        scopes[scope_id].setdefault('query', {})

                        scopes[scope_id]['query'][key] = value
            else:
                scopes.setdefault(scope_id, {})
                scopes[scope_id].setdefault('match_all', global_and_search)
                scopes[scope_id].setdefault('query', {})
                scopes[scope_id]['query'][key] = value
        else:
            # If query if empty, create an empty scope 0.
            scopes.setdefault(DEFAULT_SCOPE_ID, {})
            scopes[DEFAULT_SCOPE_ID].setdefault('match_all', scope_match_all)
            scopes[DEFAULT_SCOPE_ID].setdefault('query', {})

        return {
            'operators': operators, 'result_scope': result_scope,
            'scopes': scopes
        }

    def deindex_instance(self, instance):
        """
        Optional method to remove an model instance from the search index.
        """

    def get_resolved_field_map(self, search_model):
        result = {}
        for search_field in self.get_search_model_fields(search_model=search_model):
            backend_field_type = self.field_map.get(
                search_field.field_type
            )

            if backend_field_type:
                result[search_field.get_full_name()] = backend_field_type
            else:
                logger.warning(
                    'Unknown field type "%s" for model "%s"',
                    search_field.get_full_name(),
                    search_model.get_full_name()
                )

        return result

    def get_search_field_transformation(self, search_field):
        if search_field not in self.__class__._search_field_transformations:
            field_map = self.get_resolved_field_map(
                search_model=search_field.search_model
            )
            transformation = field_map[search_field.field].get(
                'transformation', SearchModel.function_return_same
            )
            self.__class__._search_field_transformations[
                search_field
            ] = transformation

        return self.__class__._search_field_transformations[search_field]

    def get_search_model_fields(self, search_model):
        result = search_model.search_fields.copy()
        result.append(
            SearchField(search_model=search_model, field='id', label='ID')
        )
        return result

    def get_status(self):
        """
        Backend specific method to provide status and statistics information.
        """
        if not hasattr(self, '_get_status'):
            return MESSAGE_FEATURE_NO_STATUS
        else:
            return self._get_status()

    def index_instance(self, instance, exclude_model=None, exclude_kwargs=None):
        """
        Optional method to add or update an model instance to the search
        index.
        """

    def index_search_model(self, search_model, range_string=None):
        """
        Optional method to add or update all instance of a model.
        """

    def initialize(self):
        if not self.__class__._initialized:
            self.__class__._initialized = True
            self._initialize()

    def _initialize(self):
        """
        Optional method to setup the backend. Executed once on every boot up.
        """

    def reset(self, search_model=None):
        """
        Optional method to clear all search indices.
        """

    def search(
        self, query, search_model, user, global_and_search=False
    ):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )

        result = self.decode_query(
            global_and_search=global_and_search, query=query
        )
        # Recursive call to the backend's search using queries as unscoped
        # and then merge then using the corresponding operator.
        queryset = self.solve_scope(
            operators=result['operators'],
            result_scope=result['result_scope'], search_model=search_model,
            scopes=result['scopes'], user=user
        )

        if search_model.permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=search_model.permission, queryset=queryset,
                user=user
            )

        return SearchBackend.limit_queryset(queryset=queryset)

    def solve_scope(
        self, search_model, user, result_scope, scopes, operators
    ):
        if len(scopes) > 1:
            ignore_limit = True
        else:
            ignore_limit = False

        try:
            # Try scopes.
            scope = scopes[result_scope]
        except KeyError:
            try:
                # Try operators.
                operator = operators[result_scope]
            except KeyError:
                raise DynamicSearchException(
                    'Scope `{}` not found.'.format(result_scope)
                )
            else:
                result = None
                for scope in operator['scopes']:
                    queryset = self.solve_scope(
                        operators=operators, result_scope=scope,
                        search_model=search_model, scopes=scopes, user=user
                    )

                    if result is None:
                        result = queryset
                    else:
                        result = operator['function'](result, queryset)

                return result
        else:
            try:
                query = self.cleanup_query(
                    query=scope['query'], search_model=search_model
                )
            except KeyError:
                raise DynamicSearchException(
                    'Scope `{}` does not specify a query.'.format(result_scope)
                )
            else:
                if query:
                    return self._search(
                        global_and_search=scope['match_all'],
                        ignore_limit=ignore_limit, search_model=search_model,
                        query=query, user=user
                    )
                else:
                    return search_model.get_queryset().none()

    def tear_down(self):
        """
        Optional method to clean up and/or destroy search backend structures
        like indices.
        """

    def upgrade(self):
        """
        Optional method to upgrade the search backend persistent structures.
        """


class SearchField:
    """
    Search for terms in fields that directly belong to the parent
    SearchModel.
    """
    def __init__(
        self, search_model, field, label=None, transformation_function=None
    ):
        self.search_model = search_model
        self.field = field
        self._label = label
        self.transformation_function = transformation_function

    def __repr__(self):
        return '<{}: {}>'.format(
            self.__class__.__name__, self.field
        )

    @cached_property
    def field_type(self):
        return self.get_model_field().__class__

    def get_full_name(self):
        return self.field

    def get_model(self):
        return self.search_model.model

    def get_model_field(self):
        return get_fields_from_path(model=self.get_model(), path=self.field)[-1]

    @property
    def label(self):
        return self._label or self.get_model_field().verbose_name

    @cached_property
    def model(self):
        return self.search_model.model

    @cached_property
    def related_model(self):
        return self.get_model_field().model

    @cached_property
    def reverse_path(self):
        return reverse_field_path(model=self.model, path=self.field)[1]


class SearchModel(AppsModuleLoaderMixin):
    _loader_module_name = 'search'
    _registry = {}

    @staticmethod
    def function_return_same(value):
        return value

    @classmethod
    def all(cls):
        return sorted(
            list(set(cls._registry.values())), key=lambda x: x.label
        )

    @classmethod
    def as_choices(cls):
        return cls._registry

    @classmethod
    def get(cls, name):
        try:
            result = cls._registry[name]
        except KeyError:
            raise KeyError(_('Unknown search model `%s`.') % name)
        else:
            if getattr(result, 'serializer_path', None):
                result.serializer = import_string(
                    dotted_path=result.serializer_path
                )

        return result

    @classmethod
    def get_default(cls):
        for search_class in cls.all():
            if search_class.default:
                return search_class

    @classmethod
    def get_for_model(cls, instance):
        # Works the same for model classes and model instances.
        return cls.get(name=instance._meta.label.lower())

    @classmethod
    def get_through_models(cls):
        through_models = {}

        for search_model in cls.all():
            for related_model, related_path in search_model.get_related_models():
                # Check is each related model is connected to a many to many.
                for field in related_model._meta.get_fields():
                    if field.many_to_many:
                        try:
                            through_model = field.through
                        except AttributeError:
                            through_model = field.remote_field.through

                        through_models.setdefault(through_model, {})
                        through_models[through_model].setdefault(related_model, set())
                        through_models[through_model][related_model].add(related_path)

        return through_models

    def __init__(
        self, app_label, model_name, default=False, label=None,
        list_mode=None, manager_name=None, permission=None,
        queryset=None, serializer_path=None
    ):
        self.default = default
        self._label = label
        self.app_label = app_label
        self.list_mode = list_mode or LIST_MODE_CHOICE_LIST
        self.model_name = model_name.lower()
        self._proxies = []  # Lazy
        self.permission = permission
        self.queryset = queryset
        self.search_fields = []
        self.search_fields_dict = {}
        self.serializer_path = serializer_path

        self.add_model_field(field='id', label=_('ID'))

        self.manager_name = manager_name or self.model._meta.default_manager.name

        if default:
            for search_class in self.__class__._registry.values():
                search_class.default = False

        self.__class__._registry[self.get_full_name()] = self

    def __repr__(self):
        return '<{}: {}>'.format(
            self.__class__.__name__, self.label
        )

    def __str__(self):
        return force_text(s=self.label)

    def add_model_field(self, *args, **kwargs):
        """
        Add a search field that directly belongs to the parent SearchModel.
        """
        search_field = SearchField(self, *args, **kwargs)
        self.search_fields.append(search_field)
        self.search_fields_dict[search_field.get_full_name()] = search_field

    def add_proxy_model(self, app_label, model_name):
        model_name = model_name.lower()
        self._proxies.append(
            {
                'app_label': app_label, 'model_name': model_name
            }
        )

        self.__class__._registry['{}.{}'.format(app_label, model_name)] = self

    @cached_property
    def fields_direct(self):
        result = []

        for search_field in self.search_fields:
            field_name = search_field.field

            if '__' not in field_name:
                result.append(search_field)

        return result

    @cached_property
    def fields_related(self):
        result = []

        for search_field in self.search_fields:
            field_name = search_field.field

            if '__' in field_name:
                result.append(search_field)

        return result

    def get_fields_simple_list(self):
        """
        Returns a list of the fields for the SearchModel.
        """
        result = []
        for search_field in self.search_fields:
            result.append(
                (search_field.get_full_name(), search_field.label)
            )

        return sorted(result, key=lambda x: x[1])

    def get_full_name(self):
        return '{}.{}'.format(self.app_label, self.model_name)

    def get_queryset(self):
        if self.queryset is not None:
            return self.queryset()
        else:
            return self.model._meta.managers_map[self.manager_name].all()

    def get_related_models(self):
        result = set()
        for search_field in self.search_fields:
            obj, path = reverse_field_path(
                model=self.model, path=search_field.field
            )
            if path:
                # Ignore search model fields.
                result.add((obj, path))

        return result

    def get_search_field(self, full_name):
        try:
            return self.search_fields_dict[full_name]
        except KeyError:
            raise KeyError('No search field named: %s' % full_name)

    @cached_property
    def label(self):
        if not self._label:
            self._label = self.model._meta.verbose_name
        return self._label

    @cached_property
    def base_model(self):
        return self.model._meta.proxy_for_model or self.model

    @cached_property
    def model(self):
        return apps.get_model(
            app_label=self.app_label, model_name=self.model_name
        )

    @cached_property
    def pk(self):
        return self.get_full_name()

    def populate(
        self, backend, instance, exclude_model=None, exclude_kwargs=None
    ):
        result = {}
        search_model = SearchModel.get_for_model(instance=instance)

        for direct_field in search_model.fields_direct:
            field_name = direct_field.field

            # Fetch the fields that produce a finite number of results.
            value = getattr(instance, field_name)

            result[field_name] = backend.get_search_field_transformation(
                search_field=direct_field
            )(value)

        for related_field in search_model.fields_related:
            # Fetch the fields that produce an undetermined number of
            # results.
            # This implementation is a balance between memory usage and
            # performance.
            # Attempting to fetch all of these fields at the same time using
            # joins creates a single query that kills the database and
            # causes the OOM to kill this method's process due to run away
            # memory usage.
            # This implementation is not infinitely scalable but has been
            # tested up to 500 pages per document version + 500 pages per
            # document file with an indexing chunk size of 25 for a total
            # of 25,000 page results per super call.
            field_name = related_field.field
            last_field = field_name.split('__')[-1]

            sub_queryset = related_field.related_model._meta.default_manager.filter(
                **{
                    '{}'.format(related_field.reverse_path): instance.pk
                }
            ).values_list(last_field, flat=True)

            if exclude_model and related_field.related_model == exclude_model:
                sub_queryset = sub_queryset.exclude(**exclude_kwargs)

            final_value = []
            for value in sub_queryset.distinct():
                final_value.append(
                    backend.get_search_field_transformation(
                        search_field=related_field
                    )(value) or ''
                )

            result[field_name] = ' '.join(final_value)

        return result

    @property
    def proxies(self):
        result = []
        for proxy in self._proxies:
            result.append(
                apps.get_model(
                    app_label=proxy['app_label'], model_name=proxy['model_name']
                )
            )
        return result



import logging

from django.apps import apps
from django.db.utils import OperationalError, ProgrammingError
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.classes import PropertyHelper
from mayan.apps.templating.classes import Template

from .exceptions import WorkflowStateActionError

__all__ = ('WorkflowAction',)
logger = logging.getLogger(name=__name__)


class DocumentStateHelper(PropertyHelper):
    @staticmethod
    @property
    def constructor(*args, **kwargs):
        return DocumentStateHelper(*args, **kwargs)

    def get_result(self, name):
        return self.instance.workflows.get(workflow__internal_name=name)


class WorkflowActionMetaclass(type):
    _registry = {}

    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(
            mcs, name, bases, attrs
        )

        if not new_class.__module__ == __name__:
            mcs._registry[
                '{}.{}'.format(new_class.__module__, name)
            ] = new_class

        return new_class


class WorkflowAction(
    AppsModuleLoaderMixin, metaclass=WorkflowActionMetaclass
):
    _loader_module_name = 'workflow_actions'
    fields = {}
    previous_dotted_paths = ()

    @classmethod
    def load_modules(cls):
        super().load_modules()

        for action_class in WorkflowAction.get_all():
            action_class.migrate()

    @classmethod
    def clean(cls, request, form_data=None):
        return form_data

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def get_all(cls):
        return sorted(cls._registry.values(), key=lambda x: x.label)

    @classmethod
    def get_choices(cls):
        apps_name_map = {
            app.name: app for app in apps.get_app_configs()
        }

        # Match each workflow action to an app.
        apps_workflow_action_map = {}

        for klass in WorkflowAction.get_all():
            for app_name, app in apps_name_map.items():
                if klass.__module__.startswith(app_name):
                    apps_workflow_action_map.setdefault(app, [])
                    apps_workflow_action_map[app].append(
                        (klass.id(), klass.label)
                    )

        result = [
            (app.verbose_name, workflow_actions) for app, workflow_actions in apps_workflow_action_map.items()
        ]

        # Sort by app, then by workflow action.
        return sorted(result, key=lambda x: (x[0], x[1]))

    @classmethod
    def id(cls):
        return '{}.{}'.format(cls.__module__, cls.__name__)

    @classmethod
    def migrate(cls):
        WorkflowStateAction = apps.get_model(
            app_label='document_states', model_name='WorkflowStateAction'
        )
        for previous_dotted_path in cls.previous_dotted_paths:
            try:
                WorkflowStateAction.objects.filter(
                    action_path=previous_dotted_path
                ).update(action_path=cls.id())
            except (OperationalError, ProgrammingError):
                # Ignore errors during the database migration and
                # quit further attempts.
                return

    def __init__(self, form_data=None):
        self.form_data = form_data

    def execute(self, context):
        raise NotImplementedError

    def get_fields(self):
        return getattr(self, 'fields', {})

    def get_field_order(self):
        return getattr(self, 'field_order', ())

    def get_media(self):
        return getattr(self, 'media', {})

    def get_form_schema(self, workflow_state, request=None):
        result = {
            'fields': self.get_fields(),
            'media': self.get_media(),
            'widgets': self.get_widgets(),
        }

        field_order = self.get_field_order()

        if field_order:
            result['field_order'] = field_order

        return result

    def get_widgets(self):
        return getattr(self, 'widgets', {})

    def render_field(self, field_name, context):
        try:
            result = Template(
                template_string=self.form_data.get(field_name, '')
            ).render(
                context=context
            )
        except Exception as exception:
            raise WorkflowStateActionError(
                _('%(field_name)s template error: %(exception)s') % {
                    'field_name': field_name, 'exception': exception
                }
            )

        logger.debug('%s template result: %s', field_name, result)

        return result



from django import forms

from mayan.apps.views.forms import DetailForm

from .models import UserLocaleProfile


class LocaleProfileForm(forms.ModelForm):
    class Meta:
        fields = ('language', 'timezone')
        model = UserLocaleProfile
        widgets = {
            'language': forms.Select(
                attrs={
                    'class': 'select2'
                }
            ),
            'timezone': forms.Select(
                attrs={
                    'class': 'select2'
                }
            )
        }


class LocaleProfileForm_view(DetailForm):
    class Meta:
        fields = ('language', 'timezone')
        model = UserLocaleProfile



from mayan.apps.messaging.events import event_message_created
from mayan.apps.messaging.models import Message
from mayan.apps.storage.events import event_download_file_created
from mayan.apps.storage.models import DownloadFile

from ..events import event_document_version_exported
from ..literals import (
    DOCUMENT_FILE_ACTION_PAGES_NEW, DOCUMENT_FILE_ACTION_PAGES_APPEND,
    DOCUMENT_FILE_ACTION_PAGES_KEEP, DOCUMENT_VERSION_EXPORT_MESSAGE_SUBJECT
)

from .base import GenericDocumentTestCase


class DocumentVersionTestCase(GenericDocumentTestCase):
    def test_version_new_file_new_pages(self):
        test_document_version_page_content_objects = self.test_document_version.page_content_objects

        self.assertEqual(self.test_document.versions.count(), 1)

        self._upload_test_document_file(action=DOCUMENT_FILE_ACTION_PAGES_NEW)

        self.assertEqual(self.test_document.versions.count(), 2)

        self.assertNotEqual(
            self.test_document_version.page_content_objects,
            test_document_version_page_content_objects
        )
        self.assertEqual(
            self.test_document_version.page_content_objects,
            list(self.test_document.file_latest.pages.all())
        )

    def test_version_new_version_keep_pages(self):
        test_document_version_page_content_objects = self.test_document_version.page_content_objects

        self.assertEqual(self.test_document.versions.count(), 1)

        self._upload_test_document_file(action=DOCUMENT_FILE_ACTION_PAGES_KEEP)

        self.assertEqual(self.test_document.versions.count(), 1)

        self.assertEqual(
            self.test_document_version.page_content_objects,
            test_document_version_page_content_objects
        )
        self.assertNotEqual(
            self.test_document_version.page_content_objects,
            list(self.test_document.file_latest.pages.all())
        )

    def test_version_new_file_append_pages(self):
        test_document_version_page_content_objects = self.test_document_version.page_content_objects

        self.assertEqual(self.test_document.versions.count(), 1)
        self.assertEqual(self.test_document.files.count(), 1)

        self._upload_test_document_file(action=DOCUMENT_FILE_ACTION_PAGES_APPEND)

        self.assertEqual(self.test_document.files.count(), 2)
        self.assertEqual(self.test_document.versions.count(), 2)

        test_document_version_expected_page_content_objects = list(
            self.test_document.files.first().pages.all()
        )
        test_document_version_expected_page_content_objects.extend(
            list(
                self.test_document.files.last().pages.all()
            )
        )

        self.assertNotEqual(
            self.test_document_version.page_content_objects,
            test_document_version_page_content_objects
        )
        self.assertEqual(
            self.test_document_version.page_content_objects,
            test_document_version_expected_page_content_objects
        )

    def test_method_get_absolute_url(self):
        self.assertTrue(self.test_document.version_active.get_absolute_url())


class DocumentVersionExportModelTestCase(GenericDocumentTestCase):
    def test_document_version_export(self):
        self._create_test_user()

        download_file_count = DownloadFile.objects.count()

        self._clear_events()

        self.test_user.locale_profile.language = 'es'

        self.test_document_version.export_to_download_file(
            user=self.test_user
        )

        self.assertEqual(
            DownloadFile.objects.count(), download_file_count + 1
        )

        test_download_file = DownloadFile.objects.first()
        test_message = Message.objects.first()

        self.assertNotEqual(
            test_message.subject, DOCUMENT_VERSION_EXPORT_MESSAGE_SUBJECT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 3)

        self.assertEqual(events[0].action_object, self.test_document_version)
        self.assertEqual(events[0].actor, self.test_user)
        self.assertEqual(events[0].target, test_download_file)
        self.assertEqual(events[0].verb, event_download_file_created.id)

        self.assertEqual(events[1].action_object, test_download_file)
        self.assertEqual(events[1].actor, self.test_user)
        self.assertEqual(events[1].target, self.test_document_version)
        self.assertEqual(events[1].verb, event_document_version_exported.id)

        self.assertEqual(events[2].action_object, None)
        self.assertEqual(events[2].actor, test_message)
        self.assertEqual(events[2].target, test_message)
        self.assertEqual(events[2].verb, event_message_created.id)



from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..permissions import permission_workflow_template_view

from .mixins.workflow_runtime_proxy_mixins import (
    WorkflowRuntimeProxyStateViewTestMixin, WorkflowRuntimeProxyViewTestMixin
)
from .mixins.workflow_template_mixins import WorkflowTemplateTestMixin


class WorkflowRuntimeProxyViewTestCase(
    WorkflowRuntimeProxyViewTestMixin, WorkflowTemplateTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._create_test_document_stub()

    def test_workflow_runtime_proxy_document_list_view_no_permission(self):
        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_document_list_view()
        self.assertNotContains(
            response=response, text=self._test_workflow_template.label,
            status_code=404
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_runtime_proxy_document_list_view_with_workflow_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_document_list_view()
        self.assertContains(
            response=response, text=self._test_workflow_template.label,
            status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_runtime_proxy_document_list_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_document_list_view()
        self.assertNotContains(
            response=response, text=self._test_workflow_template.label,
            status_code=404
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_runtime_proxy_document_list_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_document_list_view()
        self.assertContains(
            response=response, text=self._test_workflow_template.label,
            status_code=200
        )
        self.assertContains(
            response=response, text=self.test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_workflow_runtime_proxy_document_list_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_document_list_view()
        self.assertContains(
            response=response, text=self._test_workflow_template.label,
            status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_runtime_proxy_list_view_no_permission(self):
        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_list_view()
        self.assertNotContains(
            response=response, text=self._test_workflow_template.label,
            status_code=200
        )

    def test_workflow_runtime_proxy_list_view_with_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_list_view()
        self.assertContains(
            response=response, text=self._test_workflow_template.label,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class WorkflowRuntimeProxyStateViewTestCase(
    WorkflowRuntimeProxyStateViewTestMixin, WorkflowTemplateTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._create_test_document_stub()

    def test_workflow_runtime_proxy_state_document_list_view_no_permission(self):
        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_state_document_list_view()
        self.assertNotContains(
            response=response,
            text=self._test_workflow_template_states[0].label, status_code=404
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=404
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_runtime_proxy_state_document_list_view_with_workflow_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_state_document_list_view()
        self.assertContains(
            response=response,
            text=self._test_workflow_template_states[0].label, status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_runtime_proxy_state_document_list_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_state_document_list_view()
        self.assertNotContains(
            response=response,
            text=self._test_workflow_template_states[0].label, status_code=404
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_runtime_proxy_state_document_list_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_state_document_list_view()
        self.assertContains(
            response=response,
            text=self._test_workflow_template_states[0].label, status_code=200
        )
        self.assertContains(
            response=response, text=self.test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_workflow_runtime_proxy_state_document_list_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_state_document_list_view()
        self.assertContains(
            response=response,
            text=self._test_workflow_template_states[0].label, status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_runtime_proxy_state_list_view_no_permission(self):
        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_state_list_view()
        self.assertNotContains(
            response=response,
            text=self._test_workflow_template_states[0].label, status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_runtime_proxy_state_list_view_with_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_runtime_proxy_state_list_view()
        self.assertContains(
            response=response,
            text=self._test_workflow_template_states[0].label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.document_states.permissions import permission_workflow_template_edit
from mayan.apps.document_states.tests.base import ActionTestCase
from mayan.apps.document_states.tests.mixins.workflow_template_mixins import WorkflowTemplateTestMixin
from mayan.apps.document_states.tests.mixins.workflow_template_state_mixins import WorkflowTemplateStateActionViewTestMixin
from mayan.apps.testing.tests.base import GenericViewTestCase

from ..models import Tag
from ..workflow_actions import AttachTagAction, RemoveTagAction

from .mixins import TagTestMixin


class TagWorkflowActionTestCase(TagTestMixin, ActionTestCase):
    def setUp(self):
        super().setUp()
        self._create_test_tag()

    def test_tag_attach_action(self):
        action = AttachTagAction(form_data={'tags': Tag.objects.all()})
        action.execute(context={'document': self.test_document})

        self.assertEqual(self.test_tag.documents.count(), 1)
        self.assertTrue(self.test_document in self.test_tag.documents.all())

    def test_tag_remove_action(self):
        self.test_tag.attach_to(document=self.test_document)

        action = RemoveTagAction(form_data={'tags': Tag.objects.all()})
        action.execute(context={'document': self.test_document})

        self.assertEqual(self.test_tag.documents.count(), 0)


class TagWorkflowActionViewTestCase(
    WorkflowTemplateStateActionViewTestMixin, WorkflowTemplateTestMixin,
    GenericViewTestCase
):
    def test_tag_attach_action_create_view(self):
        self._create_test_workflow_template()
        self._create_test_workflow_template_state()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        response = self._request_test_workflow_template_state_action_create_post_view(
            class_path='mayan.apps.tags.workflow_actions.AttachTagAction'
        )
        self.assertEqual(response.status_code, 302)

    def test_tag_remove_action_create_view(self):
        self._create_test_workflow_template()
        self._create_test_workflow_template_state()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        response = self._request_test_workflow_template_state_action_create_post_view(
            class_path='mayan.apps.tags.workflow_actions.RemoveTagAction'
        )
        self.assertEqual(response.status_code, 302)



import logging

from django.apps import apps
from django.contrib.auth.models import Group
from django.db import models
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.classes import EventManagerSave
from mayan.apps.events.decorators import method_event
from mayan.apps.user_management.permissions import permission_group_view

from .classes import Permission
from .events import event_role_created, event_role_edited
from .managers import RoleManager, StoredPermissionManager

logger = logging.getLogger(name=__name__)


class Role(ExtraDataModelMixin, models.Model):
    """
    This model represents a Role. Roles are permission units. They are the
    only object to which permissions can be granted. They are themselves
    containers too, containing Groups, which are organization units. Roles
    are the basic method to grant a permission to a group. Permissions granted
    to a group using a role, are granted for the entire system.
    """
    label = models.CharField(
        help_text=_('A short text describing the role.'), max_length=128,
        unique=True, verbose_name=_('Label')
    )
    permissions = models.ManyToManyField(
        related_name='roles', to='StoredPermission',
        verbose_name=_('Permissions')
    )
    groups = models.ManyToManyField(
        related_name='roles', to=Group, verbose_name=_('Groups')
    )

    objects = RoleManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _('Role')
        verbose_name_plural = _('Roles')

    def __str__(self):
        return self.label

    def get_absolute_url(self):
        return reverse(viewname='permissions:role_list')

    def get_group_count(self, user):
        """
        Return the numeric count of groups that have this role contains.
        The count is filtered by access.
        """
        return self.get_groups(user=user).count()

    def get_groups(self, user):
        """
        Return a filtered queryset groups that have this role contains.
        """
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )

        return AccessControlList.objects.restrict_queryset(
            permission=permission_group_view, queryset=self.groups.all(),
            user=user
        )
    get_group_count.short_description = _('Group count')

    def get_permission_count(self):
        """
        Return the numeric count of permissions that have this role
        has granted. The count is filtered by access.
        """
        return self.permissions.count()
    get_permission_count.short_description = _('Permission count')

    def grant(self, permission):
        self.permissions.add(permission.stored_permission)

    def groups_add(self, queryset, _event_actor=None):
        for obj in queryset:
            self.groups.add(obj)
            event_role_edited.commit(
                action_object=obj, actor=_event_actor or self._event_actor,
                target=self
            )

    def groups_remove(self, queryset, _event_actor=None):
        for obj in queryset:
            self.groups.remove(obj)
            event_role_edited.commit(
                action_object=obj, actor=_event_actor or self._event_actor,
                target=self
            )

    def natural_key(self):
        return (self.label,)
    natural_key.dependencies = ['auth.Group', 'permissions.StoredPermission']

    def permissions_add(self, queryset, _event_actor=None):
        for obj in queryset:
            self.permissions.add(obj)
            event_role_edited.commit(
                action_object=obj, actor=_event_actor or self._event_actor,
                target=self
            )

    def permissions_remove(self, queryset, _event_actor=None):
        for obj in queryset:
            self.permissions.remove(obj)
            event_role_edited.commit(
                action_object=obj, actor=_event_actor or self._event_actor,
                target=self
            )

    def revoke(self, permission):
        self.permissions.remove(permission.stored_permission)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_role_created,
            'target': 'self',
        },
        edited={
            'event': event_role_edited,
            'target': 'self',
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class StoredPermission(models.Model):
    """
    This model is the counterpart of the permissions.classes.Permission
    class. Allows storing a database counterpart of a permission class.
    It is used to store the permissions help by a role or in an ACL.
    """
    namespace = models.CharField(max_length=64, verbose_name=_('Namespace'))
    name = models.CharField(max_length=64, verbose_name=_('Name'))

    objects = StoredPermissionManager()

    class Meta:
        ordering = ('namespace',)
        unique_together = ('namespace', 'name')
        verbose_name = _('Permission')
        verbose_name_plural = _('Permissions')

    def __str__(self):
        try:
            return force_text(s=self.volatile_permission)
        except KeyError:
            return self.name

    def get_absolute_url(self):
        return reverse(
            viewname='permissions:stored_permission_detail', kwargs={
                'stored_permission_id': self.pk
            }
        )

    @cached_property
    def volatile_permission_id(self):
        """
        Return the identifier of the real permission class represented by
        this model instance.
        """
        return '{}.{}'.format(self.namespace, self.name)

    @cached_property
    def volatile_permission(self):
        """
        Returns the real class of the permission represented by this model
        instance.
        """
        return Permission.get(
            pk=self.volatile_permission_id, class_only=True
        )

    def natural_key(self):
        return (self.namespace, self.name)

    def user_has_this(self, user):
        """
        Helper method to check if a user has been granted this permission.
        The check is done sequentially over all of the user's groups and
        roles. The check is interrupted at the first positive result.
        The check always returns True for superusers or staff users.
        """
        if user.is_superuser or user.is_staff:
            logger.debug(
                'Permission "%s" granted to user "%s" as superuser or staff',
                self, user
            )
            return True

        if not user.is_authenticated:
            return False

        if Role.objects.filter(groups__user=user, permissions=self).exists():
            return True
        else:
            logger.debug(
                'Fallthru: Permission "%s" not granted to user "%s"', self, user
            )
            return False



from django.db import models

from mayan.apps.testing.tests.base import GenericViewTestCase

from ..classes import ModelCopy
from ..permissions import permission_object_copy

from .literals import TEST_OBJECT_LABEL
from .mixins import CommonViewTestMixin, ObjectCopyViewTestMixin


class CommonViewTestCase(CommonViewTestMixin, GenericViewTestCase):
    def test_about_view(self):
        response = self._request_about_view()
        self.assertContains(response=response, text='About', status_code=200)


class ObjectCopyViewTestCase(ObjectCopyViewTestMixin, GenericViewTestCase):
    auto_create_test_object = True
    auto_create_test_object_fields = {
        'label': models.CharField(max_length=32, unique=True)
    }
    auto_create_test_object_instance_kwargs = {
        'label': TEST_OBJECT_LABEL
    }

    def setUp(self):
        super().setUp()
        ModelCopy(model=self.TestModel, register_permission=True).add_fields(
            field_names=('label',)
        )

    def test_object_copy_view_no_permission(self):
        test_object_count = self.TestModel.objects.count()
        response = self._request_object_copy_view()
        self.assertTrue(response.status_code, 404)

        queryset = self.TestModel.objects.all()
        self.assertEqual(queryset.count(), test_object_count)
        test_object_labels = queryset.values_list('label', flat=True)

        self.assertTrue(
            TEST_OBJECT_LABEL in test_object_labels
        )
        self.assertFalse(
            '{}_1'.format(TEST_OBJECT_LABEL) in test_object_labels
        )

    def test_object_copy_view_with_access(self):
        self.grant_access(obj=self.test_object, permission=permission_object_copy)

        test_object_count = self.TestModel.objects.count()
        response = self._request_object_copy_view()
        self.assertTrue(response.status_code, 302)

        queryset = self.TestModel.objects.all()
        self.assertEqual(queryset.count(), test_object_count + 1)
        test_object_labels = queryset.values_list('label', flat=True)

        self.assertTrue(
            TEST_OBJECT_LABEL in test_object_labels
        )
        self.assertTrue(
            '{}_1'.format(TEST_OBJECT_LABEL) in test_object_labels
        )



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_workflow_template_edited
from ..permissions import (
    permission_workflow_template_edit, permission_workflow_template_view
)

from .literals import TEST_WORKFLOW_TEMPLATE_STATE_LABEL
from .mixins.workflow_template_mixins import WorkflowTemplateTestMixin
from .mixins.workflow_template_state_mixins import WorkflowTemplateStateAPIViewTestMixin


class WorkflowTemplateStatesAPIViewTestCase(
    WorkflowTemplateStateAPIViewTestMixin, WorkflowTemplateTestMixin,
    BaseAPITestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_workflow_template()

    def test_workflow_template_state_create_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_workflow_template_state_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_workflow_template.refresh_from_db()
        self.assertEqual(self._test_workflow_template.states.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_template_state_create_api_view_with_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self._test_workflow_template.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template.states.first().label,
            TEST_WORKFLOW_TEMPLATE_STATE_LABEL
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(
            events[0].action_object, self._test_workflow_template_state
        )
        self.assertEqual(events[0].target, self._test_workflow_template)
        self.assertEqual(events[0].verb, event_workflow_template_edited.id)

    def test_workflow_template_state_delete_api_view_no_permission(self):
        self._create_test_workflow_template_state()

        self._clear_events()

        response = self._request_test_workflow_template_state_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_workflow_template.refresh_from_db()
        self.assertEqual(self._test_workflow_template.states.count(), 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_template_state_delete_api_view_with_access(self):
        self._create_test_workflow_template_state()

        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self._test_workflow_template.refresh_from_db()
        self.assertEqual(self._test_workflow_template.states.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].target, self._test_workflow_template)
        self.assertEqual(events[0].verb, event_workflow_template_edited.id)

    def test_workflow_template_state_detail_api_view_no_permission(self):
        self._create_test_workflow_template_state()

        self._clear_events()

        response = self._request_test_workflow_template_state_detail_api_view()

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_template_state_detail_api_view_with_access(self):
        self._create_test_workflow_template_state()

        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['id'], self._test_workflow_template_state.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_template_state_list_api_view_no_permission(self):
        self._create_test_workflow_template_state()

        self._clear_events()

        response = self._request_test_workflow_template_state_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_template_state_list_api_view_with_access(self):
        self._create_test_workflow_template_state()

        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_view
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['id'],
            self._test_workflow_template_state.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_template_state_edit_api_view_via_patch_no_permission(self):
        self._create_test_workflow_template_state()

        test_workflow_template_state_label = self._test_workflow_template_state.label

        self._clear_events()

        response = self._request_test_workflow_template_state_edit_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_workflow_template_state.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_state.label,
            test_workflow_template_state_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_template_state_edit_api_view_via_patch_with_access(self):
        self._create_test_workflow_template_state()

        test_workflow_template_state_label = self._test_workflow_template_state.label

        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_edit_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_workflow_template_state.refresh_from_db()
        self.assertNotEqual(
            self._test_workflow_template_state.label,
            test_workflow_template_state_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(
            events[0].action_object, self._test_workflow_template_state
        )
        self.assertEqual(events[0].target, self._test_workflow_template)
        self.assertEqual(events[0].verb, event_workflow_template_edited.id)

    def test_workflow_template_state_edit_api_view_via_put_no_permission(self):
        self._create_test_workflow_template_state()

        test_workflow_template_state_label = self._test_workflow_template_state.label

        self._clear_events()

        response = self._request_test_workflow_template_state_edit_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_workflow_template_state.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_state.label,
            test_workflow_template_state_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_template_state_edit_api_view_via_put_with_access(self):
        self._create_test_workflow_template_state()

        test_workflow_template_state_label = self._test_workflow_template_state.label

        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        self._clear_events()

        response = self._request_test_workflow_template_state_edit_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_workflow_template_state.refresh_from_db()
        self.assertNotEqual(
            self._test_workflow_template_state.label,
            test_workflow_template_state_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(
            events[0].action_object, self._test_workflow_template_state
        )
        self.assertEqual(events[0].target, self._test_workflow_template)
        self.assertEqual(events[0].verb, event_workflow_template_edited.id)



# Generated by Django 2.2.12 on 2020-06-15 07:41

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ErrorLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, verbose_name='Internal name')),
            ],
            options={
                'verbose_name': 'Error log',
                'verbose_name_plural': 'Error logs',
                'ordering': ('name',),
            },
        ),
        migrations.CreateModel(
            name='ErrorLogPartition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=128, verbose_name='Internal name')),
                ('error_log', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='partitions', to='logging.ErrorLog', verbose_name='Error log')),
            ],
            options={
                'verbose_name': 'Error log partition',
                'verbose_name_plural': 'Error log partitions',
                'unique_together': {('error_log', 'name')},
            },
        ),
        migrations.CreateModel(
            name='ErrorLogPartitionEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('datetime', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Date and time')),
                ('text', models.TextField(blank=True, null=True, verbose_name='Text')),
                ('error_log_partition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entries', to='logging.ErrorLogPartition', verbose_name='Error log partition')),
            ],
            options={
                'verbose_name': 'Error log partition entry',
                'verbose_name_plural': 'Error log partition entries',
                'ordering': ('datetime',),
                'get_latest_by': 'datetime',
            },
        ),
    ]



from .email_backends import *  # NOQA
from .sane_scanner_backends import *  # NOQA
from .staging_folder_backends import *  # NOQA
from .watch_folder_backends import *  # NOQA
from .web_form_backends import *  # NOQA



from ...source_backends.email_backends import (
    SourceBackendIMAPEmail, SourceBackendPOP3Email
)
from ...source_backends.literals import (
    DEFAULT_EMAIL_IMAP_MAILBOX, DEFAULT_EMAIL_IMAP_SEARCH_CRITERIA,
    DEFAULT_EMAIL_IMAP_STORE_COMMANDS, DEFAULT_EMAIL_METADATA_ATTACHMENT_NAME,
    DEFAULT_EMAIL_POP3_TIMEOUT, DEFAULT_PERIOD_INTERVAL,
    SOURCE_UNCOMPRESS_CHOICE_NEVER
)

from ..literals import (
    TEST_EMAIL_ATTACHMENT_AND_INLINE, TEST_SOURCE_BACKEND_EMAIL_PATH
)

from .base_mixins import SourceTestMixin, SourceViewTestMixin


class EmailSourceBackendTestMixin(SourceTestMixin):
    _create_source_method = '_create_test_email_source_backend'
    _test_email_source_content = None

    def _create_test_email_source_backend(self, extra_data=None):
        backend_data = {
            'document_type_id': self.test_document_type.pk,
            'host': '',
            'interval': DEFAULT_PERIOD_INTERVAL,
            'metadata_attachment_name': DEFAULT_EMAIL_METADATA_ATTACHMENT_NAME,
            'password': '',
            'port': '',
            'ssl': True,
            'store_body': False,
            'username': '',
            '_test_content': TEST_EMAIL_ATTACHMENT_AND_INLINE
        }

        if extra_data:
            backend_data.update(extra_data)

        self._create_test_source(
            backend_path=TEST_SOURCE_BACKEND_EMAIL_PATH,
            backend_data=backend_data
        )


class IMAPEmailSourceTestMixin(SourceTestMixin):
    _create_source_method = '_create_test_imap_email_source'

    def _create_test_imap_email_source(self, extra_data=None):
        backend_data = {
            'document_type_id': self.test_document_type.pk,
            'execute_expunge': True,
            'host': '',
            'interval': DEFAULT_PERIOD_INTERVAL,
            'mailbox': DEFAULT_EMAIL_IMAP_MAILBOX,
            'mailbox_destination': '',
            'metadata_attachment_name': DEFAULT_EMAIL_METADATA_ATTACHMENT_NAME,
            'password': '',
            'port': '',
            'search_criteria': DEFAULT_EMAIL_IMAP_SEARCH_CRITERIA,
            'ssl': True,
            'store_body': False,
            'store_commands': DEFAULT_EMAIL_IMAP_STORE_COMMANDS,
            'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER,
            'username': ''
        }

        if extra_data:
            backend_data.update(extra_data)

        self._create_test_source(
            backend_path=SourceBackendIMAPEmail.get_class_path(),
            backend_data=backend_data
        )


class POP3EmailSourceTestMixin(SourceTestMixin):
    _create_source_method = '_create_test_pop3_email_source'

    def _create_test_pop3_email_source(self, extra_data=None):
        backend_data = {
            'document_type_id': self.test_document_type.pk,
            'host': '',
            'interval': DEFAULT_PERIOD_INTERVAL,
            'metadata_attachment_name': DEFAULT_EMAIL_METADATA_ATTACHMENT_NAME,
            'password': '',
            'port': '',
            'ssl': True,
            'store_body': False,
            'timeout': DEFAULT_EMAIL_POP3_TIMEOUT,
            'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER,
            'username': ''
        }

        if extra_data:
            backend_data.update(extra_data)

        self._create_test_source(
            backend_path=SourceBackendPOP3Email.get_class_path(),
            backend_data=backend_data
        )


class EmailSourceBackendViewTestMixin(SourceViewTestMixin):
    def _request_test_email_source_create_view(self, extra_data=None):
        data = {
            'document_type_id': self.test_document_type.pk,
            'host': '127.0.0.1',
            'interval': DEFAULT_PERIOD_INTERVAL,
            'metadata_attachment_name': DEFAULT_EMAIL_METADATA_ATTACHMENT_NAME,
            'port': '0',
            'ssl': True,
            'store_body': False,
            'username': 'username'
        }

        if extra_data:
            data.update(extra_data)

        return self._request_test_source_create_view(
            backend_path=TEST_SOURCE_BACKEND_EMAIL_PATH, extra_data=data
        )



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0074_auto_20201130_0341'),
    ]

    operations = [
        migrations.DeleteModel(
            name='DuplicatedDocumentOld',
        ),
    ]



from django.db.models import Q

from ...models.favorite_document_models import FavoriteDocument


class FavoriteDocumentAPIViewTestMixin:
    def _request_test_favorite_document_create_api_view(self):
        pk_list = list(FavoriteDocument.valid.values_list('pk', flat=True))

        response = self.post(
            viewname='rest_api:favoritedocument-list', data={
                'document_id': self.test_document.pk
            }
        )

        try:
            self.test_favorite_document = FavoriteDocument.valid.get(
                ~Q(pk__in=pk_list)
            )
        except FavoriteDocument.DoesNotExist:
            self.test_favorite_document = None

        return response

    def _request_test_favorite_document_delete_api_view(self):
        return self.delete(
            viewname='rest_api:favoritedocument-detail', kwargs={
                'favorite_document_id': self.test_favorite_document.pk
            }
        )

    def _request_test_favorite_document_detail_api_view(self):
        return self.get(
            viewname='rest_api:favoritedocument-detail', kwargs={
                'favorite_document_id': self.test_favorite_document.pk
            }
        )

    def _request_test_favorite_document_list_api_view(self):
        return self.get(viewname='rest_api:favoritedocument-list')


class FavoriteDocumentTestMixin:
    def _test_document_favorite_add(self, user=None):
        user = user or self._test_case_user

        self.test_favorite_document = FavoriteDocument.valid.add_for_user(
            document=self.test_document, user=user
        )


class FavoriteDocumentsViewTestMixin:
    def _request_test_document_favorite_add_view(self):
        return self.post(
            viewname='documents:document_favorite_add',
            kwargs={'document_id': self.test_document.pk}
        )

    def _request_test_document_favorites_list_view(self):
        return self.get(
            viewname='documents:document_favorite_list',
        )

    def _request_test_document_favorite_remove_view(self):
        return self.post(
            viewname='documents:document_favorite_remove',
            kwargs={'document_id': self.test_document.pk}
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.permissions import permission_acl_edit, permission_acl_view
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import menu_object, menu_tools
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.navigation.classes import SourceColumn

from .classes import DefinedStorage
from .events import event_download_file_downloaded
from .links import (
    link_download_file_delete, link_download_file_download,
    link_download_file_list
)


class StorageApp(MayanAppConfig):
    app_namespace = 'storage'
    app_url = 'storage'
    has_tests = True
    name = 'mayan.apps.storage'
    verbose_name = _('Storage')

    def ready(self):
        super().ready()
        DefinedStorage.load_modules()

        DownloadFile = self.get_model(model_name='DownloadFile')

        EventModelRegistry.register(model=DownloadFile)

        ModelEventType.register(
            model=DownloadFile, event_types=(
                event_download_file_downloaded,
            )
        )

        ModelPermission.register(
            model=DownloadFile, permissions=(
                permission_acl_edit, permission_acl_view
            )
        )

        SourceColumn(
            attribute='datetime', is_identifier=True, include_label=True,
            is_sortable=True,
            source=DownloadFile
        )
        SourceColumn(
            attribute='label', include_label=True, is_sortable=True,
            source=DownloadFile
        )
        SourceColumn(
            attribute='content_object', include_label=True,
            label=_('Source object'), source=DownloadFile,
            is_attribute_absolute_url=True,
        )
        SourceColumn(
            attribute='filename', include_label=True, is_sortable=True,
            source=DownloadFile
        )

        menu_object.bind_links(
            links=(
                link_download_file_delete, link_download_file_download,
            ), sources=(DownloadFile,)
        )
        menu_tools.bind_links(links=(link_download_file_list,))



from io import StringIO
from unittest import skip

from django.core import management
from django.db import models

from mayan.apps.testing.tests.base import BaseTestCase
from mayan.apps.testing.tests.utils import mute_stdout

from ..classes import SearchBackend, SearchModel

from .mixins import SearchTestMixin


class SearchIndexObjectManagementCommandTestCaseMixin(SearchTestMixin):
    auto_create_test_object_model = True
    auto_create_test_object_fields = {
        'test_field': models.CharField(max_length=8)
    }

    def _call_command(self, id_range_string):
        with mute_stdout():
            management.call_command(
                'search_index_objects',
                self._test_model_search.get_full_name(), id_range_string
            )

    def _do_search(self, search_terms):
        return self.search_backend.search(
            search_model=self._test_model_search,
            query={
                'test_field': search_terms
            }, user=self._test_case_user
        )

    def _setup_test_model_search(self):
        self._test_model_search = SearchModel(
            app_label=self.TestModel._meta.app_label,
            model_name=self.TestModel._meta.model_name,
        )
        self._test_model_search.add_model_field(field='test_field')

    def test_command_calling(self):
        self._create_test_object(instance_kwargs={'test_field': 'abc'})

        backend = SearchBackend.get_instance()
        backend.reset()

        self._call_command(id_range_string=self.test_objects[0].pk)

    def test_command_artifacts(self):
        self._create_test_object(instance_kwargs={'test_field': 'abc'})
        self._create_test_object(instance_kwargs={'test_field': 'xyz'})

        queryset = self._do_search(
            search_terms=self.test_objects[0].test_field
        )
        self.assertTrue(self.test_objects[0] in queryset)

        queryset = self._do_search(
            search_terms=self.test_objects[1].test_field
        )
        self.assertTrue(self.test_objects[1] in queryset)

        backend = SearchBackend.get_instance()
        backend.reset()

        queryset = self._do_search(
            search_terms=self.test_objects[0].test_field
        )
        self.assertTrue(self.test_objects[0] not in queryset)

        queryset = self._do_search(
            search_terms=self.test_objects[1].test_field
        )
        self.assertTrue(self.test_objects[1] not in queryset)

        self._call_command(id_range_string=self.test_objects[0].pk)

        queryset = self._do_search(
            search_terms=self.test_objects[0].test_field
        )
        self.assertTrue(self.test_objects[0] in queryset)

        queryset = self._do_search(
            search_terms=self.test_objects[1].test_field
        )
        self.assertTrue(self.test_objects[1] not in queryset)

        self._call_command(id_range_string=self.test_objects[1].pk)

        queryset = self._do_search(
            search_terms=self.test_objects[0].test_field
        )
        self.assertTrue(self.test_objects[0] in queryset)

        queryset = self._do_search(
            search_terms=self.test_objects[1].test_field
        )
        self.assertTrue(self.test_objects[1] in queryset)


class DjangoSearchIndexObjectManagementCommandTestCase(
    SearchIndexObjectManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'
    """Test against Django backend."""

    @skip('Backend does not support indexing.')
    def test_command_artifacts(self):
        """Backend does not support indexing."""


@skip('Skip until a Mock ElasticSearch server class is added.')
class ElasticSearchIndexObjectManagementCommandTestCase(
    SearchIndexObjectManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.elasticsearch.ElasticSearchBackend'
    """Test against ElasticSearch backend."""


class WhooshSearchIndexObjectManagementCommandTestCase(
    SearchIndexObjectManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend'
    """Test against Whoosh backend."""


class SearchStatusManagementCommandTestCaseMixin(SearchTestMixin):
    _ignore_result = False
    auto_create_test_object_model = True
    auto_create_test_object_fields = {
        'test_field': models.CharField(max_length=8)
    }

    def _do_search(self, search_terms):
        return self.search_backend.search(
            search_model=self._test_model_search,
            query={
                'test_field': search_terms
            }, user=self._test_case_user
        )

    def _setup_test_model_search(self):
        self._test_model_search = SearchModel(
            app_label=self.TestModel._meta.app_label,
            model_name=self.TestModel._meta.model_name,
        )
        self._test_model_search.add_model_field(field='test_field')

    def test_command(self):
        backend = SearchBackend.get_instance()
        backend.reset()

        self._create_test_object(instance_kwargs={'test_field': 'abc'})
        self._create_test_object(instance_kwargs={'test_field': 'xyz'})

        queryset = self._do_search(
            search_terms=self.test_objects[0].test_field
        )
        self.assertTrue(self.test_objects[0] in queryset)

        output = StringIO()
        options = {
            'stdout': output
        }

        management.call_command(command_name='search_status', **options)
        output.seek(0)

        count = 0

        lines = output.readlines()

        for line in lines:
            if self._test_model_search.model_name in line.lower():
                model_name, count = line.split(':')
                count = int(count)

        self.assertEqual(count, len(self.test_objects))


class DjangoSearchStatusManagementCommandTestCase(
    SearchStatusManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'
    """Test against DjangoSearch backend."""


@skip('Skip until a Mock ElasticSearch server class is added.')
class ElasticSearchStatusManagementCommandTestCase(
    SearchStatusManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.elasticsearch.ElasticSearchBackend'
    """Test against ElasticSearch backend."""


class WhooshSearchStatusManagementCommandTestCase(
    SearchStatusManagementCommandTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend'
    """Test against WhooshSearch backend."""



class ConvertError(Exception):
    """
    Base exception for all coverter app exceptions
    """


class AppImageError(ConvertError):
    """
    Exception to allow app specific error codes. These are errors that might
    need additional handling and that are not just a generic unknown format
    error.
    """
    def __init__(self, error_name):
        self.error_name = error_name
        super().__init__()

    def __str__(self):
        return('Error name: {}'.format(repr(self.error_name)))


class UnknownFileFormat(ConvertError):
    """
    Raised when the converter backend can't understand a file
    """


class UnkownConvertError(ConvertError):
    """
    Raised when an error is found but there is no disernible way to
    identify the kind of error
    """


class OfficeConversionError(ConvertError):
    """
    Used to encapsulate errors while executing LibreOffice or when
    LibreOffice is now available.
    """


class InvalidOfficeFormat(ConvertError):
    """
    Raised by the file type introspection code to signal that the file is not
    a office format file and that LibreOffice will not be used to process
    it.
    """


class PageCountError(ConvertError):
    """
    Raised when an error is encountered while determining the page count of a
    file.
    """


class LayerError(ConvertError):
    """
    Raise by the layer class when attempting to create a transformation in
    a layer for which it was not registered.
    """



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('sources', '0025_delete_sourcelog'),
    ]

    operations = [
        migrations.AddField(
            model_name='source',
            name='backend_data',
            field=models.TextField(blank=True, verbose_name='Backend data'),
        ),
        migrations.AddField(
            model_name='source',
            name='backend_path',
            field=models.CharField(default='', help_text='The dotted Python path to the backend class.', max_length=128, verbose_name='Backend path'),
            preserve_default=False,
        ),
    ]



from io import StringIO
import mock

from django.core import management

import mayan

from mayan.apps.testing.tests.base import BaseTestCase

from ..utils import (
    MESSAGE_NOT_LATEST, MESSAGE_UNKNOWN_VERSION, MESSAGE_UP_TO_DATE
)


class ShowVersionManagementCommandTestCase(BaseTestCase):
    create_test_case_user = False

    def test_version_command_base(self):
        out = StringIO()
        management.call_command(command_name='showversion', stdout=out)
        self.assertIn(mayan.__version__, out.getvalue())

    def test_version_command_build_string(self):
        out = StringIO()
        management.call_command(
            command_name='showversion', build_string=True, stdout=out
        )
        self.assertIn(mayan.__build_string__, out.getvalue())


class CheckVersionManagementCommandTestCase(BaseTestCase):
    def _call_command(self):
        out = StringIO()
        management.call_command(
            command_name='checkversion', stdout=out
        )
        return out.getvalue()

    @mock.patch('mayan.apps.dependencies.utils.PyPIClient.get_versions', autospec=True)
    def test_check_version_not_latest_version(self, mock_package_releases):
        mock_package_releases.return_value = ('0.0.0',)
        text = self._call_command()
        self.assertTrue(text.startswith(MESSAGE_NOT_LATEST[:-2]))

    @mock.patch('mayan.apps.dependencies.utils.PyPIClient.get_versions', autospec=True)
    def test_check_version_unknown_version(self, mock_package_releases):
        mock_package_releases.return_value = None
        text = self._call_command()
        self.assertTrue(text.startswith(MESSAGE_UNKNOWN_VERSION))

    @mock.patch('mayan.apps.dependencies.utils.PyPIClient.get_versions', autospec=True)
    def test_check_version_correct_version(self, mock_package_releases):
        mock_package_releases.return_value = (mayan.__version__,)
        text = self._call_command()
        self.assertTrue(text.startswith(MESSAGE_UP_TO_DATE))



from django.utils.translation import ugettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_c

queue_events = CeleryQueue(
    label=_('Events'), name='events', transient=True,
    worker=worker_c
)

queue_events.add_task_type(
    dotted_path='mayan.apps.events.tasks.task_event_queryset_clear',
    label=_('Clear event querysets'), name='task_event_queryset_clear',
)
queue_events.add_task_type(
    dotted_path='mayan.apps.events.tasks.task_event_queryset_export',
    label=_('Export event querysets'), name='task_event_queryset_export',
)



from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..events import (
    event_document_comment_created, event_document_comment_deleted,
    event_document_comment_edited
)
from ..models import Comment
from ..permissions import (
    permission_document_comment_create, permission_document_comment_delete,
    permission_document_comment_edit, permission_document_comment_view
)

from .mixins import DocumentCommentTestMixin, DocumentCommentViewTestMixin


class DocumentCommentViewTestCase(
    DocumentCommentViewTestMixin, DocumentCommentTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_user()
        self._create_test_document_stub()

    def test_comment_create_view_no_permission(self):
        comment_count = Comment.objects.count()

        self._clear_events()

        response = self._request_test_comment_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Comment.objects.count(), comment_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_create_view_with_permissions(self):
        comment_count = Comment.objects.count()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_comment_create
        )

        self._clear_events()

        response = self._request_test_comment_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Comment.objects.count(), comment_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document_comment)
        self.assertEqual(events[0].verb, event_document_comment_created.id)

    def test_trashed_document_comment_create_view_with_permissions(self):
        comment_count = Comment.objects.count()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_comment_create
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_comment_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Comment.objects.count(), comment_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_delete_view_no_permission(self):
        self._create_test_comment()

        comment_count = Comment.objects.count()

        self._clear_events()

        response = self._request_test_comment_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Comment.objects.count(), comment_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_delete_view_with_access(self):
        self._create_test_comment()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_comment_delete
        )

        comment_count = Comment.objects.count()

        self._clear_events()

        response = self._request_test_comment_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Comment.objects.count(), comment_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document)
        self.assertEqual(events[0].verb, event_document_comment_deleted.id)

    def test_trashed_document_comment_delete_view_with_access(self):
        self._create_test_comment()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_comment_delete
        )

        self.test_document.delete()

        comment_count = Comment.objects.count()

        self._clear_events()

        response = self._request_test_comment_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Comment.objects.count(), comment_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_detail_view_no_permission(self):
        self._create_test_comment()

        self._clear_events()

        response = self._request_test_comment_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_detail_view_with_access(self):
        self._create_test_comment()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_comment_view
        )

        self._clear_events()

        response = self._request_test_comment_detail_view()
        self.assertContains(
            response=response, text=self.test_document_comment.text,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_comment_detail_view_with_access(self):
        self._create_test_comment()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_comment_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_comment_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_edit_view_no_permission(self):
        self._create_test_comment()

        comment_text = self.test_document_comment.text

        self._clear_events()

        response = self._request_test_comment_edit_view()
        self.assertEqual(response.status_code, 404)

        self.test_document_comment.refresh_from_db()
        self.assertEqual(self.test_document_comment.text, comment_text)
        self.assertEqual(self.test_document_comment.user, self.test_user)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_edit_view_with_access(self):
        self._create_test_comment()

        self.grant_access(
            obj=self.test_document, permission=permission_document_comment_edit
        )

        comment_text = self.test_document_comment.text

        self._clear_events()

        response = self._request_test_comment_edit_view()
        self.assertEqual(response.status_code, 302)

        self.test_document_comment.refresh_from_db()
        self.assertNotEqual(self.test_document_comment.text, comment_text)
        self.assertEqual(self.test_document_comment.user, self.test_user)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document_comment)
        self.assertEqual(events[0].verb, event_document_comment_edited.id)

    def test_trashed_document_comment_edit_view_with_access(self):
        self._create_test_comment()

        self.grant_access(
            obj=self.test_document, permission=permission_document_comment_edit
        )

        comment_text = self.test_document_comment.text

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_comment_edit_view()
        self.assertEqual(response.status_code, 404)

        self.test_document_comment.refresh_from_db()
        self.assertEqual(self.test_document_comment.text, comment_text)
        self.assertEqual(self.test_document_comment.user, self.test_user)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_list_view_with_no_permission(self):
        self._create_test_comment()

        self._clear_events()

        response = self._request_test_comment_list_view()
        self.assertNotContains(
            response=response, text=self.test_document_comment.text,
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_list_view_with_access(self):
        self._create_test_comment()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_comment_view
        )

        self._clear_events()

        response = self._request_test_comment_list_view()
        self.assertContains(
            response=response, text=self.test_document_comment.text,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_comment_list_view_with_access(self):
        self._create_test_comment()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_comment_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_comment_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.http import HttpResponseRedirect
from django.utils.deprecation import MiddlewareMixin

from mayan.apps.appearance.settings import setting_ajax_redirection_code


class AjaxRedirect(MiddlewareMixin):
    def process_request(self, request):
        ajax_referer = request.META.get('HTTP_X_ALT_REFERER')

        if ajax_referer:
            request.META['HTTP_REFERER'] = ajax_referer

        return None

    def process_response(self, request, response):
        if request.is_ajax():
            if type(response) == HttpResponseRedirect:
                response.status_code = setting_ajax_redirection_code.value
        return response



from django.http import Http404
from django.utils.encoding import force_text

from .classes import SearchModel
from .literals import SEARCH_MODEL_NAME_KWARG


class SearchModelViewMixin:
    def dispatch(self, *args, **kwargs):
        self.search_model = self.get_search_model()
        return super().dispatch(*args, **kwargs)

    def get_search_model_name(self):
        return self.kwargs.get(
            SEARCH_MODEL_NAME_KWARG, self.request.GET.get(
                '_{}'.format(SEARCH_MODEL_NAME_KWARG), self.request.POST.get(
                    '_{}'.format(SEARCH_MODEL_NAME_KWARG)
                )
            )
        )

    def get_search_model(self):
        try:
            return SearchModel.get(name=self.get_search_model_name())
        except KeyError as exception:
            raise Http404(force_text(s=exception))



from .tasks import (
    task_unverify_key_signatures,
    task_verify_missing_embedded_signature,
    task_verify_key_signatures
)


def handler_verify_missing_embedded_signature(sender, **kwargs):
    task_verify_missing_embedded_signature.delay()


def handler_unverify_key_signatures(sender, **kwargs):
    task_unverify_key_signatures.apply_async(
        kwargs={
            'key_id': kwargs['instance'].key_id
        }
    )


def handler_verify_key_signatures(sender, **kwargs):
    task_verify_key_signatures.apply_async(
        kwargs={
            'key_pk': kwargs['instance'].pk
        }
    )



from django.contrib.auth import get_user_model

from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.metadata.permissions import permission_document_metadata_edit
from mayan.apps.metadata.tests.mixins import MetadataTypeTestMixin
from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import event_user_created, event_user_edited
from ..permissions import (
    permission_group_edit, permission_user_create, permission_user_delete,
    permission_user_edit, permission_user_view
)

from .mixins import (
    CurrentUserViewTestMixin, GroupTestMixin, UserGroupViewTestMixin,
    UserViewTestMixin
)


class CurrentUserViewTestCase(CurrentUserViewTestMixin, GenericViewTestCase):
    def test_current_user_detail_view_no_permission(self):
        self._clear_events()

        response = self._request_current_user_details_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_current_user_edit_get_view_no_permission(self):
        self._clear_events()

        response = self._request_current_user_edit_view()
        self.assertEqual(response.status_code, 200)

        self.assertNotContains(
            response=response, status_code=200, text='Is active'
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_current_user_edit_get_view_with_access(self):
        self.grant_access(
            obj=self._test_case_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_current_user_edit_view()
        self.assertEqual(response.status_code, 200)

        self.assertContains(
            response=response, status_code=200, text='Is active'
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_current_user_edit_post_view_no_permission(self):
        test_first_name = self._test_case_user.first_name
        test_username = self._test_case_user.username

        self._clear_events()

        response = self._request_current_user_post_view()
        self.assertEqual(response.status_code, 302)

        self._test_case_user.refresh_from_db()

        self.assertEqual(self._test_case_user.username, test_username)
        self.assertNotEqual(self._test_case_user.first_name, test_first_name)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_case_user)
        self.assertEqual(events[0].verb, event_user_edited.id)

    def test_current_user_edit_post_view_with_access(self):
        self.grant_access(
            obj=self._test_case_user, permission=permission_user_edit
        )

        test_first_name = self._test_case_user.first_name
        test_username = self._test_case_user.username

        self._clear_events()

        response = self._request_current_user_post_view()
        self.assertEqual(response.status_code, 302)

        self._test_case_user.refresh_from_db()

        self.assertNotEqual(self._test_case_user.username, test_username)
        self.assertNotEqual(self._test_case_user.first_name, test_first_name)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_case_user)
        self.assertEqual(events[0].verb, event_user_edited.id)


class CurrentUserSuperUserViewTestCase(
    CurrentUserViewTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._test_case_user.is_superuser = True
        self._test_case_user.save()

    def test_current_user_detail_view_no_permission(self):
        self._clear_events()

        response = self._request_current_user_details_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_current_user_edit_get_view_no_permission(self):
        self._clear_events()

        response = self._request_current_user_edit_view()
        self.assertEqual(response.status_code, 200)

        self.assertContains(
            response=response, status_code=200, text='Is active'
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_current_user_edit_post_view_no_permission(self):
        test_first_name = self._test_case_user.first_name
        test_username = self._test_case_user.username

        self._clear_events()

        response = self._request_current_user_post_view()
        self.assertEqual(response.status_code, 302)

        self._test_case_user.refresh_from_db()

        self.assertNotEqual(self._test_case_user.username, test_username)
        self.assertNotEqual(self._test_case_user.first_name, test_first_name)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_case_user)
        self.assertEqual(events[0].verb, event_user_edited.id)


class SuperUserViewTestCase(
    UserViewTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_superuser()

    def test_superuser_delete_view_with_access(self):
        superuser_count = get_user_model().objects.filter(
            is_superuser=True
        ).count()
        self.grant_access(
            obj=self.test_superuser, permission=permission_user_delete
        )

        self._clear_events()

        response = self._request_test_superuser_delete_view()
        self.assertEqual(response.status_code, 404)
        self.assertEqual(
            get_user_model().objects.filter(is_superuser=True).count(),
            superuser_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_superuser_detail_view_with_access(self):
        self.grant_access(
            obj=self.test_superuser, permission=permission_user_view
        )

        self._clear_events()

        response = self._request_test_superuser_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_superuser_normal_user_detail_view_with_access(self):
        self.grant_access(
            obj=self.test_superuser, permission=permission_user_view
        )

        self.test_user = self.test_superuser

        self._clear_events()

        response = self._request_test_user_detail_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class UserViewTestCase(UserViewTestMixin, GenericViewTestCase):
    def test_user_create_view_no_permission(self):
        user_count = get_user_model().objects.count()

        self._clear_events()

        response = self._request_test_user_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(get_user_model().objects.count(), user_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_create_view_with_permission(self):
        self.grant_permission(permission=permission_user_create)

        user_count = get_user_model().objects.count()

        self._clear_events()

        response = self._request_test_user_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(get_user_model().objects.count(), user_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_user)
        self.assertEqual(events[0].verb, event_user_created.id)

    def test_user_delete_view_no_permission(self):
        self._create_test_user()

        user_count = get_user_model().objects.count()

        self._clear_events()

        response = self._request_test_user_delete_single_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(get_user_model().objects.count(), user_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_delete_view_with_access(self):
        self._create_test_user()

        user_count = get_user_model().objects.count()

        self.grant_access(
            obj=self.test_user, permission=permission_user_delete
        )

        self._clear_events()

        response = self._request_test_user_delete_single_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(get_user_model().objects.count(), user_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_delete_multiple_view_no_permission(self):
        self._create_test_user()

        user_count = get_user_model().objects.count()

        self._clear_events()

        response = self._request_test_user_delete_multiple_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(get_user_model().objects.count(), user_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_delete_multiple_view_with_access(self):
        self._create_test_user()

        user_count = get_user_model().objects.count()

        self.grant_access(
            obj=self.test_user, permission=permission_user_delete
        )

        self._clear_events()

        response = self._request_test_user_delete_multiple_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(get_user_model().objects.count(), user_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_edit_view_no_permission(self):
        self._create_test_user()

        username = self.test_user.username

        self._clear_events()

        response = self._request_test_user_edit_view()
        self.assertEqual(response.status_code, 404)

        self.test_user.refresh_from_db()
        self.assertEqual(self.test_user.username, username)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_edit_view_with_access(self):
        self._create_test_user()

        username = self.test_user.username

        self.grant_access(
            obj=self.test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_user_edit_view()
        self.assertEqual(response.status_code, 302)

        self.test_user.refresh_from_db()
        self.assertNotEqual(self.test_user.username, username)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_user)
        self.assertEqual(events[0].verb, event_user_edited.id)

    def test_user_list_view_no_permission(self):
        self._create_test_user()

        self._clear_events()

        response = self._request_test_user_list_view()
        self.assertNotContains(
            response=response, text=self.test_user.username, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_list_view_with_permission(self):
        self._create_test_user()

        self.grant_access(
            obj=self.test_user, permission=permission_user_view
        )

        self._clear_events()

        response = self._request_test_user_list_view()
        self.assertContains(
            response=response, text=self.test_user.username, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class UserGroupViewTestCase(
    GroupTestMixin, UserGroupViewTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_group()
        self._create_test_user()

    def test_user_group_add_remove_get_view_no_permission(self):
        self.test_group.user_set.add(self.test_user)

        self._clear_events()

        response = self._request_test_user_group_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self.test_group),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self.test_user),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_group_add_remove_get_view_with_group_access(self):
        self.test_group.user_set.add(self.test_user)

        self.grant_access(
            obj=self.test_group, permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_user_group_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self.test_group),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self.test_user),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_group_add_remove_get_view_with_user_access(self):
        self.test_group.user_set.add(self.test_user)

        self.grant_access(
            obj=self.test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_user_group_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self.test_group),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self.test_user),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_group_add_remove_get_view_with_full_access(self):
        self.test_group.user_set.add(self.test_user)

        self.grant_access(
            obj=self.test_group, permission=permission_group_edit
        )
        self.grant_access(
            obj=self.test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_user_group_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self.test_group),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self.test_user),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_group_add_view_no_permission(self):
        self._clear_events()

        response = self._request_test_user_group_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_group not in self.test_user.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_group_add_view_with_group_access(self):
        self.grant_access(
            obj=self.test_group, permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_user_group_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_group not in self.test_user.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_group_add_view_with_user_access(self):
        self.grant_access(
            obj=self.test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_user_group_add_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self.test_group not in self.test_user.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_group_add_view_with_full_access(self):
        self.grant_access(
            obj=self.test_group, permission=permission_group_edit
        )
        self.grant_access(
            obj=self.test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_user_group_add_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self.test_group in self.test_user.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_group)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_user)
        self.assertEqual(events[0].verb, event_user_edited.id)

    def test_user_group_remove_view_no_permission(self):
        self.test_group.user_set.add(self.test_user)

        self._clear_events()

        response = self._request_test_user_group_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_group in self.test_user.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_group_remove_view_with_group_access(self):
        self.test_group.user_set.add(self.test_user)

        self.grant_access(
            obj=self.test_group, permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_user_group_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_group in self.test_user.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_group_remove_view_with_user_access(self):
        self.test_group.user_set.add(self.test_user)

        self.grant_access(
            obj=self.test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_user_group_remove_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self.test_group in self.test_user.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_group_remove_view_with_full_access(self):
        self.test_group.user_set.add(self.test_user)

        self.grant_access(
            obj=self.test_group, permission=permission_group_edit
        )
        self.grant_access(
            obj=self.test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_user_group_remove_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self.test_group not in self.test_user.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_group)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_user)
        self.assertEqual(events[0].verb, event_user_edited.id)


class UserOptionsViewTestCase(
    UserViewTestMixin, GenericViewTestCase
):
    def test_user_options_view_no_permission(self):
        self._create_test_user()

        self._clear_events()

        response = self._request_test_user_options_view()
        self.assertEqual(response.status_code, 404)

        self.test_user.user_options.refresh_from_db()
        self.assertFalse(self.test_user.user_options.block_password_change)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_user_options_view_with_access(self):
        self._create_test_user()

        self.grant_access(
            obj=self.test_user, permission=permission_user_edit
        )

        self._clear_events()

        response = self._request_test_user_options_view()
        self.assertEqual(response.status_code, 302)

        self.test_user.user_options.refresh_from_db()
        self.assertTrue(self.test_user.user_options.block_password_change)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_user)
        self.assertEqual(events[0].verb, event_user_edited.id)


class MetadataLookupIntegrationTestCase(
    MetadataTypeTestMixin, GenericDocumentViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_metadata_type()
        self.test_document_type.metadata.create(
            metadata_type=self.test_metadata_type
        )

    def test_user_list_lookup_render(self):
        self.test_metadata_type.lookup = '{{ users }}'
        self.test_metadata_type.save()
        self.test_document.metadata.create(
            metadata_type=self.test_metadata_type
        )

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self.test_metadata_type,
            permission=permission_document_metadata_edit
        )

        self._clear_events()

        response = self.get(
            viewname='metadata:metadata_edit', kwargs={
                'document_id': self.test_document.pk
            }
        )
        self.assertContains(
            response=response, text='<option value="{}">{}</option>'.format(
                self._test_case_user.username, self._test_case_user.username
            ), status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import logging

from django.utils.translation import ugettext_lazy as _

from mayan.apps.databases.classes import ModelBaseBackend

__all__ = ('MailerBackend',)
logger = logging.getLogger(name=__name__)


class MailerBackend(ModelBaseBackend):
    """
    Base class for the mailing backends. This class is mainly a wrapper
    for other Django backends that adds a few metadata to specify the
    fields it needs to be instanciated at runtime.

    The fields attribute is a list of dictionaries with the format:
    {
        'name': ''  # Field internal name
        'label': ''  # Label to show to users
        'class': ''  # Field class to use. Field classes are Python dot
                       paths to Django's form fields.
        'initial': ''  # Field initial value
        'default': ''  # Default value.
    }
    """
    _loader_module_name = 'mailers'
    class_path = ''  # Dot path to the actual class that will handle the mail

    @classmethod
    def get_fields(cls):
        return getattr(cls, 'fields', {})

    @classmethod
    def get_field_order(cls):
        return getattr(cls, 'field_order', ())

    @classmethod
    def get_form_schema(cls):
        result = {
            'fields': cls.get_fields(),
            'field_order': cls.get_field_order(),
            'widgets': cls.get_widgets()
        }

        return result

    @classmethod
    def get_widgets(cls):
        return getattr(cls, 'widgets', {})


class NullBackend(MailerBackend):
    label = _('Null backend')



from django.db import models, migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0017_auto_20150714_0056'),
    ]

    operations = [
        migrations.AlterField(
            model_name='document',
            name='description',
            field=models.TextField(
                default='', verbose_name='Description', blank=True
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='document',
            name='label',
            field=models.CharField(
                default='', max_length=255, blank=True,
                help_text='The name of the document', verbose_name='Label',
                db_index=True
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='documentversion',
            name='comment',
            field=models.TextField(
                default='', verbose_name='Comment', blank=True
            ),
            preserve_default=True,
        ),
    ]



import uuid

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0044_auto_20180823_0452'),
    ]

    operations = [
        migrations.AlterField(
            model_name='document',
            name='date_added',
            field=models.DateTimeField(
                auto_now_add=True, db_index=True, help_text='The server '
                'date and time when the document was finally processed '
                'and added to the system.', verbose_name='Added'
            ),
        ),
        migrations.AlterField(
            model_name='document',
            name='deleted_date_time',
            field=models.DateTimeField(
                blank=True, help_text='The server date and time when the '
                'document was moved to the trash.', null=True,
                verbose_name='Date and time trashed'
            ),
        ),
        migrations.AlterField(
            model_name='document',
            name='description',
            field=models.TextField(
                blank=True, default='', help_text='An optional short text '
                'describing a document.', verbose_name='Description'
            ),
        ),
        migrations.AlterField(
            model_name='document',
            name='in_trash',
            field=models.BooleanField(
                db_index=True, default=False, editable=False,
                help_text='Whether or not this document is in the trash.',
                verbose_name='In trash?'
            ),
        ),
        migrations.AlterField(
            model_name='document',
            name='label',
            field=models.CharField(
                blank=True, db_index=True, default='', help_text='The name '
                'of the document.', max_length=255, verbose_name='Label'
            ),
        ),
        migrations.AlterField(
            model_name='document',
            name='language',
            field=models.CharField(
                blank=True, default=b'eng', help_text='The dominant '
                'language in the document.', max_length=8,
                verbose_name='Language'
            ),
        ),
        migrations.AlterField(
            model_name='document',
            name='uuid',
            field=models.UUIDField(
                default=uuid.uuid4, editable=False, help_text='UUID of a '
                'document, universally Unique ID. An unique identifier '
                'generated for each document.', verbose_name='UUID'
            ),
        ),
        migrations.AlterField(
            model_name='documenttype',
            name='label',
            field=models.CharField(
                help_text='The name of the document type.', max_length=32,
                unique=True, verbose_name='Label'
            ),
        ),
        migrations.AlterField(
            model_name='documentversion',
            name='checksum',
            field=models.CharField(
                blank=True, db_index=True, editable=False,
                help_text="A hash/checkdigit/fingerprint generated from "
                "the document's binary data. Only identical documents will "
                "have the same checksum.", max_length=64, null=True,
                verbose_name='Checksum'
            ),
        ),
        migrations.AlterField(
            model_name='documentversion',
            name='comment',
            field=models.TextField(
                blank=True, default='', help_text='An optional short text '
                'describing the document version.', verbose_name='Comment'
            ),
        ),
        migrations.AlterField(
            model_name='documentversion',
            name='encoding',
            field=models.CharField(
                blank=True, editable=False, help_text='The document version '
                'file encoding. binary 7-bit, binary 8-bit, text, '
                'base64, etc.', max_length=64, null=True,
                verbose_name='Encoding'
            ),
        ),
        migrations.AlterField(
            model_name='documentversion',
            name='mimetype',
            field=models.CharField(
                blank=True, editable=False, help_text='The document '
                'version\'s file mimetype. MIME types are a standard way '
                'to describe the format of a file, in this case the file '
                'format of the document. Some examples: "text/plain" '
                'or "image/jpeg". ', max_length=255, null=True,
                verbose_name='MIME type'
            ),
        ),
        migrations.AlterField(
            model_name='documentversion',
            name='timestamp',
            field=models.DateTimeField(
                auto_now_add=True, db_index=True, help_text='The server '
                'date and time when the document version was processed.',
                verbose_name='Timestamp'
            ),
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.models import Document

from ..permissions import permission_send_document_link

from .base import ObjectLinkMailView


class MailDocumentLinkView(ObjectLinkMailView):
    object_permission = permission_send_document_link
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid.all()
    success_message = _('%(count)d document link queued for email delivery')
    success_message_plural = _(
        '%(count)d document links queued for email delivery'
    )
    title = 'Email document link'
    title_plural = 'Email document links'
    title_document = 'Email link for document: %s'



import json

from mayan.apps.django_gpg.tests.literals import TEST_KEY_PRIVATE_PASSPHRASE
from mayan.apps.django_gpg.tests.mixins import KeyTestMixin
from mayan.apps.document_states.literals import WORKFLOW_ACTION_ON_ENTRY
from mayan.apps.document_states.permissions import permission_workflow_template_edit
from mayan.apps.document_states.tests.mixins.workflow_template_mixins import WorkflowTemplateTestMixin
from mayan.apps.document_states.tests.mixins.workflow_template_state_mixins import WorkflowTemplateStateActionViewTestMixin
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..models import DetachedSignature, EmbeddedSignature
from ..workflow_actions import (
    DocumentSignatureDetachedAction, DocumentSignatureEmbeddedAction
)

from .literals import (
    DOCUMENT_SIGNATURE_DETACHED_ACTION_CLASS_PATH,
    DOCUMENT_SIGNATURE_EMBEDDED_ACTION_CLASS_PATH
)


class DocumentSignatureWorkflowActionTestCase(
    KeyTestMixin, WorkflowTemplateTestMixin,
    WorkflowTemplateStateActionViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_document_signature_detached_action_create_view(self):
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        request = self._request_test_workflow_template_state_action_create_get_view(
            class_path=DOCUMENT_SIGNATURE_DETACHED_ACTION_CLASS_PATH
        )
        self.assertEqual(request.status_code, 200)

    def test_document_signature_embedded_action_create_view(self):
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        request = self._request_test_workflow_template_state_action_create_get_view(
            class_path=DOCUMENT_SIGNATURE_EMBEDDED_ACTION_CLASS_PATH
        )
        self.assertEqual(request.status_code, 200)

    def test_document_signature_detached_action(self):
        self._upload_test_document()
        self._create_test_key_private()
        signature_count = DetachedSignature.objects.count()

        action = DocumentSignatureDetachedAction(
            form_data={
                'key': self.test_key_private.pk,
                'passphrase': TEST_KEY_PRIVATE_PASSPHRASE
            }
        )
        action.execute(context={'document': self.test_document})
        self.assertNotEqual(signature_count, DetachedSignature.objects.count())

    def test_document_signature_embedded_action(self):
        self._upload_test_document()
        self._create_test_key_private()
        signature_count = EmbeddedSignature.objects.count()

        action = DocumentSignatureEmbeddedAction(
            form_data={
                'key': self.test_key_private.pk,
                'passphrase': TEST_KEY_PRIVATE_PASSPHRASE
            }
        )
        action.execute(context={'document': self.test_document})
        self.assertNotEqual(
            signature_count, EmbeddedSignature.objects.count()
        )

    def test_document_signature_detached_action_via_workflow(self):
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._create_test_key_private()

        self._test_workflow_template_states[1].actions.create(
            label='test action', when=WORKFLOW_ACTION_ON_ENTRY,
            enabled=True,
            action_path='mayan.apps.document_signatures.workflow_actions.DocumentSignatureDetachedAction',
            action_data=json.dumps(
                obj={
                    'key': self.test_key_private.pk,
                    'passphrase': TEST_KEY_PRIVATE_PASSPHRASE
                }
            ),
        )

        self._upload_test_document()
        self._test_workflow_instance = self.test_document.workflows.first()

        signature_count = DetachedSignature.objects.count()

        self._test_workflow_instance.do_transition(
            transition=self._test_workflow_template_transition
        )
        self.assertNotEqual(
            signature_count, DetachedSignature.objects.count()
        )

    def test_document_signature_embedded_action_via_workflow(self):
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._create_test_key_private()

        self._test_workflow_template_states[1].actions.create(
            label='test action', when=WORKFLOW_ACTION_ON_ENTRY,
            enabled=True,
            action_path='mayan.apps.document_signatures.workflow_actions.DocumentSignatureEmbeddedAction',
            action_data=json.dumps(
                obj={
                    'key': self.test_key_private.pk,
                    'passphrase': TEST_KEY_PRIVATE_PASSPHRASE
                }
            ),
        )

        self._upload_test_document()
        self._test_workflow_instance = self.test_document.workflows.first()

        signature_count = EmbeddedSignature.objects.count()

        self._test_workflow_instance.do_transition(
            transition=self._test_workflow_template_transition
        )
        self.assertNotEqual(
            signature_count, EmbeddedSignature.objects.count()
        )



from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_page_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..events import event_source_created
from ..models import Source
from ..permissions import permission_sources_create, permission_sources_edit

from .mixins.base_mixins import SourceViewTestMixin
from .mixins.watch_folder_source_mixins import WatchFolderSourceTestMixin


class WatchFolderSourceViewTestCase(
    SourceViewTestMixin, WatchFolderSourceTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_watch_folder_source_create_view(self):
        self.grant_permission(permission=permission_sources_create)

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_view()

        self.assertEqual(response.status_code, 302)

        self.assertEqual(Source.objects.count(), source_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_source)
        self.assertEqual(events[0].verb, event_source_created.id)

    def test_watch_folder_source_test_view_no_permission(self):
        self._copy_test_watch_folder_document()

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.objects.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_watch_folder_source_test_view_with_access(self):
        self._copy_test_watch_folder_document()

        self.grant_access(
            obj=self.test_source, permission=permission_sources_edit
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Document.objects.count(), document_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )
